"""
This code is taken from PyPi stream_service package
https://github.com/BR1py/stream_service
"""
from __future__ import absolute_import
import asyncio
import uuid
import time
import sys
import queue
import threading

from .lib import *
from .lib.frame import *
from .lib.buffer import *
from .lib.rpc import *

DEBUG = False  # Variable used for development proposes (more print outputs given)

__package__ = 'stream_service'
__version__ = '0.1.0'
__licence__ = 'MIT'
__author__ = 'B.R.'
__url__ = 'https://github.com/BR1py/stream_service'
__description__ = 'Client/Server implementation for data streaming in channels based on python >= 3.5'

# helper classes for the STRM_CLientThread

class SrvInfo():
    '''
    This class holds the server related information
    '''

    def __init__(self, c_id, control_key, RPCMethodDescriptions):
        self.c_id = c_id
        self.control_key = control_key
        self.RPCMethodDescriptions = RPCMethodDescriptions


class ClientChannel():
    '''
    ClientChannel object for each channel used (owned/writer/reader) in the STRM_Client we have such an object
    The object contains all channel related information
    '''

    def __init__(self, name, chl_id, server_maxsize=None, client_maxsize=None, full_behavior=RING_BUFFER_FULL,
                 loop=None,
                 owned_control_key=None):
        '''

        :param name: channel name
        :param chl_id: channel id
        :param queue_maxsize: maximum buffer-size of the read queue
        :param loop:  event loop  - if the loop is None the read queue will be a normal not a asynchio Queue
                                    in this case  the commands must not be awaited
        :param is_ring_buffer: If wait_if_full is False and queue is full:
                               True - oldest element will be deleted and new element will be placed in
                               False - queue will be cleared and new element is put in
        :param wait_if_full: If this is True the read queue will be blocked until the the data is taken out of the queue
                             this ensures data consistency on client side, on server side we can only ensure if the
                             buffer_size given to the channel_reader is sufficient to catch all the data.
        '''
        self._name = name
        self._id = chl_id
        self._loop = loop
        self._owned_control_key = owned_control_key
        self._server_maxsize = server_maxsize
        self._client_maxsize = client_maxsize
        self._full_behavior = full_behavior
        self._active = True

        # The queue_maxsize is increased by one to be bigger then on the server
        if loop is not None:
            if client_maxsize is not None:
                self._buffer = Buffer(client_maxsize, full_behavior, loop=loop,
                                      chl_id=chl_id, name=self.name)
            else:
                self._buffer = Buffer(None, full_behavior, loop=loop,
                                      chl_id=chl_id, name=self.name)

    # give access to some internals

    @property
    def id(self):
        return self._id

    @property
    def name(self):
        return self._name

    @property
    def control_key(self):
        return self._owned_control_key

    @property
    def is_active(self):
        return self._active

    @property
    def buffer(self):
        return self._buffer

    def set_activate(self, activate=True):
        self._active = bool(activate)

    def __repr__(self):
        if self._owned_control_key is not None:
            control_key = '***********'
        else:
            control_key = 'None'
        return 'ClientChannel(name = %s, chl_id = %s, server_maxsize = %s, client_maxsize = %s' \
               'full_behavior = %s, loop = %s, owned_control_key = %s)' % (repr(self._name),
                                                                           repr(self._id),
                                                                           repr(self._server_maxsize),
                                                                           repr(self._client_maxsize),
                                                                           repr(FULL_BEHAVIOR_STR.get(
                                                                               self._full_behavior)),
                                                                           repr(self._loop),
                                                                           control_key
                                                                           )


# Client

class StreamChannelClient_Thread(threading.Thread):
    def __init__(self, name,
                 ip_address,
                 port=None,
                 ssl=None,
                 server_authenticate=None,
                 loop=None,
                 tto_default=TTO_DEFAULT,
                 tto_para_name=TTO_DEFAULT_PARA_NAME):
        '''
        Stream Server Client Thread

        :param name: name of the client
        :param ip_address: ip address of the Stream Server
        :param port: port of the Stream Server
        :param server_authenticate: authentication string for clients t connect to the server
        :param loop: optional external event loop (if not given the asyncio default event loop will be used)
        :param tto_default: default transaction time out (after this time any transaction will be stopped)
                            the default time out is only used if no explicit time out is given
                            (tto_parameter in functions)
        :param tto_para_name:  name of the tto_parameter in rpc_functions
                               If this parameter is given it will be removed and will not be given to the rpc function.
                               But it will be considered as explicit timeout of the specific call
                               (we will not use the default timeout in this case).I
        '''
        super().__init__(name='StreamChannelClient')
        self._ = loop
        if port is None:
            port = DEFAULT_BUFFER_STREAM_MGR_PORT
        self._address = (ip_address, port)
        self._ssl = ssl
        if type(server_authenticate) is not bytes:
            raise TypeError('Given server_authenticate parameter must be of type bytes')
        self._server_authenticate = server_authenticate
        self._tto_default = tto_default
        self._tto_para_name = tto_para_name
        self._me = uuid.uuid4().hex[:8]
        self._name = str(name)
        self._address_size = 2  # default will be rewritten after server connection is made
        mask = 0
        for i in range(self._address_size):
            mask = (mask << 8) + 0xFF
        self._address_mask = mask

        self._loop = loop
        self._srv_info = None
        self._srv_rpc = None
        self._client_rpc_services = {}
        self._client_id = None
        if DEBUG: print('Client initialized')
        self._write_chls = {}
        self._read_chls = {}
        self._transactions = {}
        self._server_died = False
        self._alive = False
        self._reader = None
        self._writer = None
        self._stop_event = threading.Event()
        self._ta_id_hdl = IdsHandler(self._address_size)
        # analyse clients rpc methods
        self._rpc_methods_descs, self._rpc_methods = rpc.RPCFactory.class_init_rpc_functions(self, first_items=[
            'rpc_get_rpc_info', 'rpc_activate_write_channel'])

    #give access to some internals

    @property
    def address_size(self):
        return self._address_size

    @property
    def name(self):
        return self._name

    def run(self):
        """
        # start the the thread and the main loop for reading data

        :return:
        """
        # Start the client thread
        if DEBUG: print('Client thread started')
        if self._loop is not None:
            asyncio.set_event_loop(self._loop)
        try:
            self._loop = asyncio.get_event_loop()
        except RuntimeError:
            self._loop = asyncio.new_event_loop()
            asyncio.set_event_loop(self._loop)
        if DEBUG: self._loop.set_debug(True)
        # we try to connect to the server
        self._loop.run_until_complete(self._connect())
        # we start the read loop
        self._loop.run_until_complete(self._read_from_server_forever())
        # if the read loop is stopped the client thread will be stopped
        if DEBUG: print('Client thread stopped')

    def __get_new_transaction(self, tto=None):
        '''
        a helper function to create a new transaction object
        :param tto: give the transaction_timeout for this transaction
        :return:
        '''
        if tto is None:
            tto = self._tto_default
        return Transaction(self._loop, self._transactions, self._ta_id_hdl, self._address_size, tto,
                           self._tto_para_name)

    def wait_until_alive(self):
        '''
        helper function for the thread init
        it waits until the thread is fully started and the Client can be used
        :return:
        '''
        while not self._alive:
            time.sleep(0.5)

    # main methods:
    async def _connect(self):
        """
        connect to the server is the first thing executed after the thread is started
        :return:
        """
        transaction1 = None
        try:
            if DEBUG: print('Connection to StreamChannelServer:', self._address)
            try:
                # create the first connection to the server
                reader, writer = await asyncio.open_connection(self._address[0], self._address[1], ssl=self._ssl)
            except ConnectionRefusedError:
                raise ConnectionRefusedError('Server %s not reached!' % repr(self._address))
            self._reader = reader
            self._writer = writer
            #wait for first frame received from the server (server starts the communication)
            first_rx_frame = await read_msg(reader)
            if DEBUG: print('First response from server', first_rx_frame)
            if first_rx_frame == b'0':
                raise StreamError_Overload('Server rejected connection request, to many clients connected')
            # get the address size used on the server:
            address_size = first_rx_frame[0]
            if not type(address_size) is int:
                address_size = int.from_bytes(address_size, BO)
            if address_size > 8:
                raise ValueError(
                    'Address size told by server to large (unplausible); something wrong in server response')
            # redefine global address size
            self._address_size = address_size
            mask = 0
            for i in range(self._address_size):
                mask = (mask << 8) + 0xFF
            self._address_mask = mask
            if DEBUG: print('New address size set to:', address_size)
            # seed and key:
            seed = first_rx_frame[1:]
            if DEBUG: print('Seed received:', seed)
            raw_key = get_raw_key(seed, self._server_authenticate)  # +b'fehler'
            if DEBUG: print('Raw key:', raw_key)
            transaction1 = self.__get_new_transaction()
            tx_frame = Frame(FrameHeader(container_type=CT_CMD,
                                         coder_type=CODE_PICKLE,
                                         transaction_id=transaction1.id,
                                         command_id=0  # authentication command
                                         ),
                             raw_key,  # authentication
                             self.name,  # client_name
                             self._rpc_methods_descs,  # rpc methods
                             )
            if DEBUG: print('Connection data_frame send to server', tx_frame)
            tx_bytes = tx_frame.encode()
            if DEBUG: print('Connection data_frame send to server', tx_bytes)
            # send key to server
            await send_msg(writer, tx_bytes)

            if DEBUG: print('Waiting for server to respond...')
            rx_bytes = await read_msg(reader)
            if rx_bytes is None:
                raise StreamError_Connection("Connection to Stream Server couldn't been established")
            if DEBUG: print('Server response raw bytes', rx_bytes)
            rx_frame = Frame(FrameHeader(CT_DECODE, address_size=address_size), rx_bytes)
            if DEBUG: print('Server response decoded:', rx_frame)
            if rx_frame.header.container_type == CT_EXCEPTION:
                raise StreamError_RPC(repr(rx_frame.args))
            if rx_frame.header.transaction_id != transaction1.id:
                raise StreamError_TransactionMissmatch(
                    'The received transaction_id does not match, no connection established')
            # store the response info
            # store server related information received
            self.srv_info = SrvInfo(*rx_frame.args)
            self.client_id = self.srv_info.c_id
            # create the server RPC Factory object for the server rpc calls
            self.srv_rpc = RPCFactory(self.srv_info.RPCMethodDescriptions,
                                      FrameHeader(CT_SRV_CMD, CODE_PICKLE,
                                                  transaction_id=1,
                                                  command_id=1,
                                                  client_id=0,
                                                  channel_id=self.srv_info.c_id,
                                                  address_size=self.address_size),
                                      reader, writer, self._loop, self.__get_new_transaction)
            if DEBUG: print('Server RPC methods discovered:', self.srv_rpc._rpc_methods)
            transaction1.delete()
            return 'Server connection established'
        except:
            if transaction1 is not None:
                transaction1.delete()
            raise

    async def _read_from_server_forever(self):
        '''
        This is the main loop of the Client thread it reads the data received from the server
        The event loop is on hold (awaited until the reader contains any data received from the server)
        We process here also all RPC calls to the client received from the outside
        :return:
        '''
        loop = self._loop
        reader = self._reader
        writer = self._writer
        address_size = self._address_size
        self._alive = True
        if DEBUG: print('Client read loop started', writer, reader)
        none_cnt = 0
        rx_bytes = None
        while not self._stop_event.is_set():
            try:
                rx_bytes = await asyncio.wait_for(read_msg(reader),
                                                  timeout=5)  # await for a frame received in the reader
            except asyncio.TimeoutError:
                # From time to time we break the await in the event loop to see if the while loop should be stopped
                continue
            except:
                pass
            if rx_bytes is None:
                # server died?
                none_cnt += 1
                if none_cnt > 10:
                    # we give the server a small time to come back
                    self.server_died = True
                    self._stop_event.set()
                continue
            none_cnt = 0
            # if DEBUG: print('rx_frame', rx_frame)
            if rx_bytes == b'STOP':
                # in case of client close this trigger received to stop the loop
                #  but we continue be cause we expect the self.alive flag is set to False in this case
                # otherwise we ignore the frame for security reasons,
                # we do not like to allow this simple stop from the outside it's for internal use only
                continue
            # first analysis of received frame
            rx_header = FrameHeader(CT_DECODE, rx_bytes, address_size=address_size)
            container_type = rx_header.container_type
            rx_payload = rx_bytes[len(rx_header):]
            # react depending of the container type
            if container_type == CT_DATA or container_type == CT_DATA_END:
                chl = self._read_chls.get(rx_header.channel_id)
                if chl is None:
                    continue
                await chl.buffer.put((rx_header, rx_payload))
            elif container_type == CT_CMD:
                # call client rpc command
                try:
                    rpc_cmd = self._rpc_methods[rx_header.command_id]
                except:
                    rpc_cmd = None
                if (rpc_cmd is None):  # we do not allow the access to the authentication cmd here!
                    tx_header = FrameHeader(container_type=CT_EXCEPTION,
                                            coder_type=CODE_MARSHAL,
                                            transaction_id=rx_header.transaction_id,
                                            address_size=address_size)
                    tx_frame = Frame(tx_header, StreamError_Command('Client RPC command unknown'))
                else:
                    rx_frame = Frame(rx_header)
                    rx_frame.decode_payload(rx_payload)
                    if rx_frame.args is None:
                        args = tuple()
                    else:
                        args = rx_frame.args
                    if rx_frame.kwargs is None:
                        kwargs = {}
                    else:
                        kwargs = rx_frame.kwargs
                    back = rpc_cmd(*args, **kwargs)
                    tx_header = FrameHeader(container_type=CT_RETURN,
                                            coder_type=rx_header.coder_type,
                                            transaction_id=rx_header.transaction_id,
                                            client_id=rx_header.client_id,
                                            address_size=self.address_size)
                    if back is None:
                        tx_frame = Frame(tx_header)
                    else:
                        tx_frame = Frame(tx_header, back)
                await send_msg(writer, tx_frame.encode())
                continue
            elif container_type in {CT_RETURN, CT_EXCEPTION}:
                ta = self._transactions.get(rx_header.transaction_id)
                if ta is not None:
                    await ta.buffer.put(rx_bytes)
        if self._server_died:
            self.stop_client()

    def create_new_channel(self, name, public=True, lifetime=None, single_in=True):
        """
        create a new channel on the connected server
        and create the related client internal object related to the channel
        Note:: see the related rpc command n the server
        :param name: channel name
        :param public: is channel public or not
        :param lifetime: if not None give integer in seconds for temporary channels
        :param single_in: Do we allow multi write on the channel (other clients can write or not)
        :return:
        """
        name = str(name)  # to differentiate in the dicts the name must always be a string
        back = self.srv_rpc.create_new_channel(self.client_id, name, public, lifetime, single_in)
        if back is None:
            raise StreamError_Channel('Channel creation failed no data received from server')
        chl_id, control_key = back
        new_chl = ClientChannel(name, chl_id, owned_control_key=control_key)
        # we put two entries in the internal write channel dict to make it reachable via both keys (id and name)
        self._write_chls[name] = new_chl
        self._write_chls[chl_id] = new_chl
        return 'Channel created'

    def subscribe_read_from_channel(self, chl_id_or_name,
                                    server_buffer_size=100,
                                    server_full_behavior=RING_BUFFER_FULL,
                                    fill_rate=1,
                                    local_buffer_size=None,
                                    local_wait_if_full=True):
        """
        read subscription on the channel on the server and
        creation of the related client internal objects
        see related rpc command on the server
        :param chl_id_or_name: channel name which should be subscriped
        :param server_buffer_size: buffer size on the server
        :param server_full_behavior: what happens if buffer is full (see Buffer object)
        :param fill_rate: Integer defines how often we sample,
                  * if fill rate is 1 any channel item is put in the read buffer
                  * if fill rate is 2 any second item will be put in the read buffer
                  * if fill rate is n and n item will be put in the read buffer
        :param local_buffer_size: if given the local read buffer size can be defined
                                  if None the size is calculated based on the related to server_buffer_size
                                  this ensures that under normal conditions the local client buffer cannot overrun
        :param local_wait_if_full: Should we wait in case the local buffer is full or not
        :return:
        """
        back = self.srv_rpc.subscribe_read_from_channel(self.client_id,
                                                        chl_id_or_name,
                                                        server_buffer_size,
                                                        server_full_behavior,
                                                        fill_rate)
        if back is None:
            return False
        chl_name, chl_id = back

        if local_buffer_size is None:
            local_buffer_size = server_buffer_size + 1
        # create local ClientChannel representation object
        new_chl = ClientChannel(chl_name, chl_id,
                                local_buffer_size,
                                server_full_behavior,
                                local_wait_if_full,
                                self._loop)

        self._read_chls[chl_name] = new_chl
        self._read_chls[chl_id] = new_chl
        return 'Channel read subscription finished'

    def unsubscribe_read_from_channel(self, chl_id_or_name):
        """
        delete the subscribtion for a channel
        all related objects will be deleted
        :param chl_id_or_name: channel name or id
        :return:
        """
        back = self.srv_rpc.unsubscribe_read_from_channel(self.client_id,
                                                          chl_id_or_name)
        chl = self._read_chls[chl_id_or_name]
        del self._read_chls[chl.name]
        del self._read_chls[chl.id]
        end_header = FrameHeader(CT_DATA_END,
                                 coder_type=CODE_MARSHAL,
                                 channel_id=chl._channel_id,
                                 address_size=self._address_size)
        # to stop the loop

        chl.buffer.put_nowait(data_bytes=Frame(end_header, 'END DATA').encode())
        return 'Channel read unsubscription finished'

    def subscribe_write_to_channel(self, chl_id_or_name):
        """
        subscribe to channel as additional writer
        (see related rpc command on teh server)
        :param chl_id_or_name: channel name or id for identification
        :return:
        """
        back = self.srv_rpc.subscribe_write_to_channel(self.client_id, chl_id_or_name)
        if back is None:
            raise StreamError_Channel('Channel subscription for channel %s failed' % repr(chl_id_or_name))
        chl_name, chl_id = back
        new_chl = ClientChannel(chl_name, chl_id, self._loop)
        self._write_chls[chl_name] = new_chl
        self._write_chls[chl_id] = new_chl
        return 'Channel write subscription finished'

    def start_data_transfer(self, chl_id_or_name, coder_type):
        """
        stream data on the channel start the streaming with this command
        :param chl_id_or_name: channel name
        :param coder_type: type of encoding that should be used for the given data
                           CODER_MARSHAL;CODER_PICKLE or CODER_NUMPY (if numpy is available)
                           additionally data might be packed with gzip
        :return: FrameHeader object that should be used for the following data transfer
        """
        chl = self._write_chls[chl_id_or_name]
        chl_id = chl.id
        t_id = self.srv_rpc.start_data_transfer_on_channel(self.client_id, chl_id)
        if DEBUG: print(
            'New data stream shall be created on channel: %s, transaction_id: %s' % (repr(chl), repr(t_id)))
        header = FrameHeader(CT_DATA,
                             coder_type,
                             transaction_id=t_id,
                             counter=1,
                             channel_id=chl_id,
                             address_size=self._address_size)
        header.set_counter(0)
        return header

    def end_data_transfer(self, header):
        """
        Send the end frame to stop the data transfer
        :param header: FrameHeader used for the data transfer stream that should be stopped
        :return:
        """
        end_header = FrameHeader(CT_DATA_END,
                                 coder_type=header._coder_type,
                                 channel_id=header._channel_id,
                                 transaction_id=header._transaction_id,
                                 client_id=header._client_id,
                                 address_size=header._address_size)
        frame = Frame(end_header, 'END DATA')
        t = asyncio.run_coroutine_threadsafe(send_msg(self._writer, frame.encode()), self._loop)
        t.result()

    def put_data(self, header, item, wait_for_active=False):
        """
        stream a data package into the channel
        :param header: FrameHeader used for this stream (return of start_data_transfer())
        :param item: dat item that should be send (depending on used encoder any python object might be given here)
        :param wait_for_active: If the channel is inactive (no reader subscribed the channel) we can wait for a reader
                                or skip.
                                * True - we will wait for a reader
                                * False - we do not transfer the data we skip the transfer
        :return:
        """
        t = asyncio.run_coroutine_threadsafe(self.async_put_data(header, item, wait_for_active), self._loop)
        return t.result()

    async def async_put_data(self, header, item, wait_for_active=False):
        """
        async variant of stream a data package into the channel
        :param header: FrameHeader used for this stream (return of start_data_transfer())
        :param item: dat item that should be send (depending on used encoder any python object might be given here)
        :param wait_for_active: If the channel is inactive (no reader subscribed the channel) we can wait for a reader
                                or skip.
                                * True - we will wait for a reader
                                * False - we do not transfer the data we skip the transfer
        """
        # we expect that the counter in the given header is not used yet and we will increase after sending the frame
        if self._write_chls[header.channel_id].is_active:
            data_bytes = Frame(header, item).encode()
            await send_msg(self._writer, data_bytes)
        else:
            if wait_for_active:
                data_bytes = Frame(header, item).encode()
                while 1:
                    await asyncio.sleep(1)
                    if self._write_chls[header.channel_id].is_active:
                        await send_msg(self._writer, data_bytes)
                        break
            else:
                print('skipped', self._write_chls[header.channel_id].name)
        header.count_up()
        return header

    def get_data(self, chl_id_or_name, timeout=None):
        """
        get data from the subscribed channel
        :param chl_id_or_name: channel name
        :param timeout: optional timeout in seconds
        :return: received item object (already decoded)
        """
        chl = self._read_chls[chl_id_or_name]
        header, payload = chl.buffer.get_wait(timeout=timeout)
        frame = Frame(header)
        frame.decode_payload(payload)
        args = frame.args
        if args and len(args) == 1:
            args = args[0]
        return header, args

    async def async_get_data(self, chl_id_or_name):
        """
        async variant of get data from the subscribed channel
        :param chl_id_or_name: channel name
        :param timeout: optional timeout in seconds
        :return: received item object (already decoded)
        """

        chl = self._read_chls[chl_id_or_name]
        await asyncio.sleep(0.0001)  # give other threads possibility to be executed
        header, payload = await chl.buffer.get()
        frame = Frame(header)
        frame.decode_payload(payload)
        args = get_args(*frame.args)

        return header, args

    def get_read_channel_buffer(self, chl_id_or_name):
        '''
        delivers the read buffer directly one can use the buffer get methods to take the data out
        The object can be used for direct queue access!
        :param chl_id_or_name: name or channel id
        :return: Buffer object for channel reading
        '''
        return self._read_chls[chl_id_or_name].buffer

    def stop_client(self):
        """
        stop the client
        the client will close the server connection
        :return:
        """
        self.srv_rpc.delete_client(self.srv_info.c_id, self.srv_info.control_key)
        time.sleep(0.5)
        self._stop_event.set()
        if self._server_died:
            self.writer = None
            self.reader = None
            self._ = None
            raise StreamError_Connection('Server died.')

    # rpc
    def create_rpc_client_service(self, client_name_or_id):
        """
        create a RPCFactory object for another client to access the rpc methods in this client
        :param client_name_or_id: target client name or ip
        :return:
        """
        back = self.srv_rpc.get_client_info(self.client_id, client_name_or_id)
        #print(client_name_or_id,back)
        new_client_rpc = RPCFactory(back['RPC_METHOD_DESC'],
                                FrameHeader(CT_CMD, CODE_PICKLE,
                                            transaction_id=1,
                                            command_id=3, # we set the default to echo
                                            client_id=back['ID'],
                                            address_size=self.address_size),
                                self._reader, self._writer, self._loop, self.__get_new_transaction)
        # ToDo store internally?
        return new_client_rpc

    # Those are the standard rpc methods that must be available on each client
    # The user can extend the function by creating an own client class which inherits this class as a super class

    def rpc_activate_write_channel(self, chl_name_or_id, activate=True):
        """
        helper method used by the server to activate or deactivate the writing to a channel
        This reduces the network load because if no reader subscribed the channel no data must be send
        :param chl_name_or_id: channel name or id
        :param activate:
                        * True - channel is active
                        * False - channel is inactive
        :return:
        """
        if chl_name_or_id not in self._write_chls:
            raise StreamError_Channel('Unknown channel %s given!' % chl_name_or_id)
        chl = self._write_chls[chl_name_or_id]
        chl.set_activate(activate)
        print('Activation request received:', chl.name, activate)
        return True

    def rpc_echo(self, *args,**kwargs):
        """
        With this method one can test the rpc connection
        :param client_id: first parameter on server rpc commands is always the client id
                          (will be replaced by client object after frame is received)
        :param args: arguments list to be echoed into the return
        :param kwargs:  keyword arguments dict to be echoed into the return

        :return args,kwargs
        """
        return args,kwargs



"""
This code is taken from PyPi stream_service package
https://github.com/BR1py/stream_service
"""

#from __future__ import absolute_import
import asyncio
import random
from .helpers import *
#Possible buffer full reactions on server

# Run the buffer queue as a ringbuffer (delete oldest item and put new in)
RING_BUFFER_FULL=0 # Default behavior
# Run the buffer queue in clear mode which means if full delete all items and put new item in
CLEAR_BUFFER_FULL=1
# Run the buffer queue in skip mode which means if full the new item will be ignored and the old items are kept in the buffer queue
SKIP_BUFFER_FULL=2
# We wait endless until the buffer is free (this will block the loop!)
WAIT_BUFFER_FULL=3
# The Full Exception will be raised
RAISE_BUFFER_FULL=4
FULL_BEHAVIOR_STR={RING_BUFFER_FULL:'RING_BUFFER_FULL',CLEAR_BUFFER_FULL:'CLEAR_BUFFER_FULL',
                   SKIP_BUFFER_FULL:'SKIP_BUFFER_FULL',WAIT_BUFFER_FULL:'WAIT_BUFFER_FULL',
                   RAISE_BUFFER_FULL:'RAISE_BUFFER_FULL'}

class Buffer(asyncio.Queue):

    def __init__(self,maxsize=100,full_behavior=RING_BUFFER_FULL,fill_rate=1,loop=None,**kwargs):
        """

        Buffer object used in the server and client to store the data items after receive or before sending
        The objects behaves like a asyncio.Queue() but it is extended.
        Be sides the async calls we have also classical calls to access the queue objects.

        Especially the reaction in case the buffer is full is different and can be modified by the
        full_behavior parameter.

        :param maxsize: the maximum size of the buffer
        :param full_behavior: defines the reaction in case of QueueFull exceptions:
                              - RING_BUFFER_FULL  (=0) - (Default) The buffer is a ring buffer
                                                         Full reaction: delete the oldest item and put the new item in
                              - CLEAR_BUFFER_FULL (=1) - The buffer is a clearing buffer
                                                         Full reaction: delete all old items (clear)
                                                                        and put the new item in
                              - SKIP_BUFFER_FULL  (=2) - The buffer is a resistant buffer
                                                         Full reaction: ignore the new item keep the old ones
                              - WAIT_BUFFER_FULL  (=3) - The buffer is a blocking buffer
                                                         Full reaction: ignore the timeouts and wait until the buffer
                                                                        has place to put the item
                                                         Warning: This will block any calling loop!
                              - RAISE_BUFFER_FULL (=4) - The buffer behaves like a Queue() and raises
                                                         the QueueFull Exception

        :param fill_rate: integer (default == 1 -> no down-sampling) used to down sample the number of items put into
                          this buffer.
        :param loop: optional parameter to give the related event_loop (if not given parameter will be taken via
                     asyncio.get_event_loop() call)
        :param kwargs: Additional parameters that might be stored as attributes in the buffer object (attributes u
                       sed in the super class will be ignored
        """
        # set the additional attributes
        for k, v in kwargs.items():
            self.__setattr__(k,v)

        if loop is not None:
            super().__init__(maxsize,loop=loop)
            self._loop=loop
        else:
            self.loop=loop=asyncio.get_event_loop()
            super().__init__(maxsize,loop=loop)
        
        self._fill_rate=fill_rate
        self._fill_cnt=0
        self._full_behavior=full_behavior
        self._full_cnt=0
        self._on_full={RING_BUFFER_FULL:self.__on_full_ringbuffer,
                       CLEAR_BUFFER_FULL: self.__on_full_clearbuffer,
                       SKIP_BUFFER_FULL: self.__on_full_skipbuffer,
                       WAIT_BUFFER_FULL: self.__on_full_waitbuffer,
                       RAISE_BUFFER_FULL: self.__on_full_raise
                       }.get(full_behavior)
        if self._on_full is None:
            raise ValueError('Given full_behavior = %s is not valid'%(repr(full_behavior)))

    # give access to some internals

    @property
    def full_cnt(self):
        return self._full_cnt

    @property
    def fill_cnt(self):
        return self._fill_cnt

    # reaction methods in case the buffer is full:
    def __on_full_ringbuffer(self,item):
        super().get_nowait()
        super().put_nowait(item)
        self._full_cnt = (self._full_cnt+1)&0b11111111

    def __on_full_clearbuffer(self,item):
        self.clear()
        super().put_nowait(item)
        self._full_cnt = (self._full_cnt + 1) & 0b11111111

    def __on_full_skipbuffer(self,_):
        self._full_cnt = (self._full_cnt + 1) & 0b11111111
        return

    def __on_full_waitbuffer(self,item):
        future = asyncio.run_coroutine_threadsafe(super().put(item), self._loop)
        self._full_cnt = (self._full_cnt + 1) & 0b11111111
        return future.result()


    def __on_full_raise(self,_):
        self._full_cnt = (self._full_cnt + 1) & 0b11111111
        raise asyncio.QueueFull()

    def clear(self):
        """
        delete all items in the queue
        :return:
        """
        self._fill_cnt=self._fill_rate # by this we ensure that the next item will be put into the queue!
        while not self.empty:
            super().get_nowait()


    async def put(self,item,force=False):
        """
        async standard method for put data into the queue
        :param item: item to be put into the queue
        """
        cnt=self._fill_cnt
        cnt = (cnt+1) & 0xFFFFFFFF
        self._fill_cnt=cnt
        if force or not cnt%self._fill_rate:
            await super().put(item)
        else:
            await asyncio.sleep(0)

    async def put_wait_for(self,item,timeout,raise_except=False):
        '''
        short version of: asyncio.wait_for(self.put(item),timeout=timeout)

        Timeout reaction:
            If raise_except parameter is set the TimeoutError exception will be raised
            Else the method will react on the defined full_behavior

        :param item: item to be put in the queue
        :param timeout: timeout in seconds
        :param raise_except: True - raise TimeoutError Exception
                             False (default) - run full_behavior reaction
        '''
        try:
            return await asyncio.wait_for(self.put(item),timeout=timeout)
        except TimeoutError:
            if raise_except:
                raise
            self._on_full(item)


    async def get_wait_for(self, timeout):
        '''
        short version of: asyncio.wait_for(self.get(), timeout=timeout)

        Timeout reaction:
            TimeoutError exception will be raised

        :param timeout: timeout in seconds
        :return: item taken out of the buffer
        '''
        return await asyncio.wait_for(self.get(), timeout=timeout)

    #non async methods
    
    def put_wait(self,item,timeout,raise_except=False):
        '''
        Non async call for an await put with timeout

        ( behaves like put() in a normal queue.Queue() )

        Timeout reaction:
            If raise_except parameter is set the TimeoutError exception will be raised
            Else the method will react on the defined full_behavior

        :param item: item to be put in the queue
        :param timeout: timeout in seconds
        :param raise_except: True - raise TimeoutError Exception
                             False (default) - run full_behavior reaction
        '''
        future = asyncio.run_coroutine_threadsafe(self.put_wait_for(item,timeout,raise_except), self._loop)
        future.result()

    def put_nowait(self,item,raise_except=False,force=False):
        '''
        put a item in the queue and do not wait in case queue is full run timeout reaction

        Timeout reaction:
            If raise_except parameter is set the TimeoutError exception will be raised
            Else the method will react on the defined full_behavior

        :param item: item to be put in the queue
        :param raise_except: True - raise TimeoutError Exception
                             False (default) - run full_behavior reaction

        :param item:
        :return:
        '''
        cnt=self._fill_cnt
        cnt = (cnt+1) & 0xFFFFFFFF
        self._fill_cnt=cnt
        if force or not cnt%self._fill_rate:
            try:
                return super().put_nowait(item)
            except asyncio.QueueFull:
                if raise_except:
                    raise
                self._on_full(item)
        else:
            pass

    def get_wait(self,timeout=None):
        '''
        Non async call for an await get_wait_for() with given timeout
        ( behaves like get() in a normal queue.Queue() )

        Timeout reaction:
            TimeoutError exception will be raised

        :param timeout: timeout in seconds or None for endless waiting
        :return: item taken out of the buffer
        '''
        if timeout is None:
            future = asyncio.run_coroutine_threadsafe(self.get(), self._loop)
        else:
            future = asyncio.run_coroutine_threadsafe(self.get_wait_for(timeout), self._loop)
        return future.result()


class Transaction():
    '''
    The transaction object is mainly used in RPC calls to identify the specific call and to handle the returns
    it creates a receive buffer related to a specific rpc call
    '''

    def __init__(self,loop, transactions,ta_id_hdl, address_size, tto=TTO_DEFAULT, tto_para_name=TTO_DEFAULT_PARA_NAME):
        self.__loop = asyncio.get_event_loop()
        self.__tto = float(tto)
        self.__tto_para_name = str(tto_para_name)
        self.__transactions = transactions
        self.__ta_id_hdl=ta_id_hdl
        self.__address_size = int(address_size)
        # Constants will be overwritten in __init_transaction():
        self.__read_queue = None
        self.__t_id =ta_id= self.__ta_id_hdl.get_new_id()
        self.__buffer = Buffer(1, RAISE_BUFFER_FULL, transaction_id=ta_id)
        # enter this transaction in the transaction dict
        self.__transactions[ta_id] = self


    def set_tto(self, tto):
        self.__tto = tto

    @property
    def id(self):
        return self.__t_id

    @property
    def tto_para_name(self):
        return self.__tto_para_name

    @property
    def buffer(self):
        return self.__buffer

    def delete(self):
        del self.__transactions[self.__t_id]
        self.__ta_id_hdl.free_id(self.__t_id)



"""
In this example we create 4 Clients that will connect to the server and setup several channels
Then different type of data is streamed in between the clients


    Client1             Client3
    owns "Chl1"         owns "Chl3"
    reads "Chl2"        reads "Chl1" fill_rate=2 (takes only every second item)
    writes "Chl3"
            \          /
             \        /
              \      /
               Server
              /      \
             /        \
            /          \
    Client2             Client4
    owns "Chl2"         owns "Chl4"
    reads "Chl1"        reads "Chl3"
    reads "Chl4"        reads "Chl1"
    writes "Chl3"

REMARK: To let the example run start the run_server.py script in a parallel process

REMARK: Ensure that the root path of stream_service package is available in the python interpreter search path
"""

import asyncio
import sys
import os
import time
import threading
import ssl
import shutil

if "-root_path" in sys.argv: #started via shell script
    ROOT=os.path.dirname(os.path.dirname(os.path.normpath(sys.argv[(sys.argv.index("-root_path")+1)])))
else:
    ROOT=os.path.dirname(os.path.dirname(__file__))
if ROOT not in sys.path:
    sys.path.append(ROOT)

from stream_service import *

START_TIME=None
DEBUG=False
MULTIWRITE_LOOP=500 # Defines how many elements are written in the multi write loops for "Chl3"
USE_SSL=False
COLLECT_STATISTICS=True # Starts the server statistics channel with very high rate
CHL1_ONLY = False  # if True we just run data exchange on "Chl1"
RESULT_FOLDER='results' #name of the folder in which the resultfiles will be stored

receive_stop=threading.Event()

def stream_file(client,chl_name,file,chunk_size=1000,start_event=None):
    if start_event is not None:
        start_event.wait()
    header=client.start_data_transfer(chl_name,CODE_MARSHAL)
    with open(file,'r') as fh:
        cnt = 0
        while 1:
            data=fh.read(chunk_size)
            client.put_data(header,data,True)
            if DEBUG: print('data put on server',client.name,chl_name)
            if data is None or len(data)==0:
                client.end_data_transfer(header)
                break
            cnt += 1
    print('Client %s(Chl %s): Streaming of file %s done.'%(client.name,chl_name,repr(file)),
          '; Number of items transferred:',cnt,
          '; Duration since start:',time.time()-START_TIME )
    return True

def receive_file(client,chl_name,target_file_name,take_stats=None):
    ta_ids=set()
    with open(target_file_name,'w') as fh:
        cnt=0
        ends=0
        while 1:
            if cnt >= 200:
                if take_stats is not None:
                    take_stats.set()
            header,data=client.get_data(chl_name)
            ta_id=header.transaction_id
            if ta_id not in ta_ids:
                ta_ids.add(header.transaction_id)
            if DEBUG: print('received data on:',client.name,chl_name)#,data)
            elif header.container_type == CT_DATA_END:
                if ta_id in ta_ids:
                    ta_ids.remove(ta_id)
                if len(ta_ids)==0:
                    break
                continue
            fh.write(data)
            if receive_stop.is_set():
                break
            cnt+=1
    print('Client %s(Chl %s): file write done'%(client.name,chl_name),
          '; Number of items transferred:',cnt,
          '; Duration since start:',time.time()-START_TIME)
    return True

def multi_write(client,chl_name,start_event=None):
    if start_event is not None:
        start_event.wait()
    header=client.start_data_transfer(chl_name,CODE_MARSHAL)
    name=client.name
    for i in range(MULTIWRITE_LOOP):
        client.put_data(header,'%i,%s\n'%(i,name),True)
        if DEBUG: print('data put on server multi write %s (%s)'%(name,chl_name))
    client.end_data_transfer(header)
    print('Client %s(Chl: %s): multi write done'%(client.name,chl_name),
          '; Number of items transferred:',MULTIWRITE_LOOP,
          '; Duration since start:',time.time()-START_TIME )
    return True

def clean_last_run():
    #delete all files created during last run
    if os.path.exists(RESULT_FOLDER):
        shutil.rmtree(RESULT_FOLDER)
    os.makedirs(RESULT_FOLDER)
    # create the test_file
    with open (RESULT_FOLDER+'/'+'test.txt','w') as fh:
        for i in range(50000):
            fh.write('%i\n'%i)

if __name__ == '__main__':
    # 1. preparations:
    start=threading.Event()
    take_stats=threading.Event()
    clean_last_run()


    # 2. Start the client threads and connect to the server:
    clt1=StreamChannelClient_Thread('Client1','127.0.0.1',server_authenticate=b'STRM_Client')
    clt1.start()
    clt1.wait_until_alive()
    #We clean up the server just to be sure!
    clt1.srv_rpc.stop_statistic_channel(clt1.client_id)

    print('Client1 connected to server')

    clt2 = StreamChannelClient_Thread('Client2','127.0.0.1', server_authenticate=b'STRM_Client')
    clt2.start()
    clt2.wait_until_alive()
    print('Client2 connected to server')

    clt3 = StreamChannelClient_Thread('Client3','127.0.0.1', server_authenticate=b'STRM_Client')
    clt3.start()
    clt3.wait_until_alive()
    print('Client3 connected to server')

    clt4 = StreamChannelClient_Thread('Client4','127.0.0.1', server_authenticate=b'STRM_Client')
    clt4.start()
    clt4.wait_until_alive()
    print('Client4 connected to server')


    # 3. create the channels
    clt1.create_new_channel('Chl1')
    print('Client1 Chl1 created')
    clt2.create_new_channel('Chl2')
    print('Client2 Chl2 created')
    clt3.create_new_channel('Chl3',single_in=False)
    print('Client3 Chl3 (multi write) created')
    clt4.create_new_channel('Chl4')
    print('Client4 Chl4 created')

    print('All clients and channels created')

    # 4. subscribe all the readers and writers to the channels:
    print('Channel subscription follows')
    clt1.subscribe_read_from_channel('Chl2')
    clt1.subscribe_write_to_channel('Chl3')

    clt2.subscribe_read_from_channel('Chl1')
    clt2.subscribe_read_from_channel('Chl4')
    clt2.subscribe_write_to_channel('Chl3')

    clt3.subscribe_read_from_channel('Chl1',fill_rate=2)

    clt4.subscribe_read_from_channel('Chl1')
    clt4.subscribe_read_from_channel('Chl3',server_buffer_size=600)

    #get the channel statistics
    if COLLECT_STATISTICS:
        stat_chl=clt1.srv_rpc.start_statistic_channel(clt1.client_id,1) #put every sampling into statistics file
        clt1.subscribe_read_from_channel(stat_chl,server_buffer_size=10000)
        print('Read also:', stat_chl)

    print('All Channel subscribed')

    # 5. create data writing and reading threads which will create the activity on the server at least:

    print('Setup all the streaming threads')
    #We first setup the readers just to be sure not to miss a package:
    read_client2_chl1 = threading.Thread(target=receive_file, args=(clt2, 'Chl1', RESULT_FOLDER+'/'+'client2_chl1.txt'))
    read_client3_chl1 = threading.Thread(target=receive_file,args=(clt3,'Chl1', RESULT_FOLDER+'/'+'client3_chl1.txt'))
    read_client4_chl1 = threading.Thread(target=receive_file,args=(clt4,'Chl1', RESULT_FOLDER+'/'+'client4_chl1.txt'))
    if not CHL1_ONLY:
        read_client1_chl2 = threading.Thread(target=receive_file, args=(clt1, 'Chl2', RESULT_FOLDER+'/'+'client1_chl2.txt'))
        read_client2_chl4 = threading.Thread(target=receive_file, args=(clt2, 'Chl4', RESULT_FOLDER+'/'+'client2_chl4.txt'))
        read_client4_chl3 = threading.Thread(target=receive_file, args=(clt4, 'Chl3', RESULT_FOLDER+'/'+'client4_multi_chl3.txt',take_stats))

    #writers (owned)
    write_client1_chl1 = threading.Thread(target=stream_file, args=(clt1, 'Chl1', RESULT_FOLDER+'/'+'test.txt',1000,start))
    if not CHL1_ONLY:
        write_client2_chl2 = threading.Thread(target=stream_file, args=(clt2, 'Chl2', RESULT_FOLDER+'/'+'test.txt',1000,start))
        write_client4_chl4 = threading.Thread(target=stream_file, args=(clt4, 'Chl4', RESULT_FOLDER+'/'+'test.txt',1000,start))
        #multi write
        write_client1_chl3 = threading.Thread(target=multi_write, args=(clt1, 'Chl3',start))
        write_client2_chl3 = threading.Thread(target=multi_write, args=(clt2, 'Chl3',start))
        write_client3_chl3 = threading.Thread(target=multi_write, args=(clt3, 'Chl3',start))

    # 6. before we start streaming we start the readers (we like to get all data and we do not want to read only a part)

    read_client2_chl1.start()
    read_client3_chl1.start()
    read_client4_chl1.start()
    print('Stream data read on Chl1 threads started')

    if COLLECT_STATISTICS:
        read_statistics = threading.Thread(target=receive_file,
                                           args=(clt1, stat_chl, RESULT_FOLDER+'/'+'statistics.txt'))
        read_statistics.start()

    if not CHL1_ONLY:
        read_client1_chl2.start()
        read_client2_chl4.start()
        read_client4_chl3.start()
        print('Stream data read on other channels threads started')

    # 7. We start the send streams now but they will wait for the start event
    if not CHL1_ONLY:
        print('Start streaming multi write on Chl3,Chl2,Chl4 threads started')
        # multi write
        write_client1_chl3.start()
        write_client2_chl3.start()
        write_client3_chl3.start()
        #other channels
        write_client2_chl2.start()
        write_client4_chl4.start()

    print('Streaming data from Client1 to Chl1 thread started')
    write_client1_chl1.start()

    # 8. start the streaming by setting the start event!
    START_TIME = time.time()
    start.set()
    print('Start-Event set at:',START_TIME)

    # 9. wait for the streaming to be finished

    #depending on processor speed one  may adapt this timigs so that the statitics taken under high load
    read_client2_chl1.join()
    read_client3_chl1.join()
    read_client4_chl1.join()
    print('Chl1 reader threads finished')

    if not CHL1_ONLY:
        read_client1_chl2.join()
        read_client2_chl4.join()
        read_client4_chl3.join() # This is the slowest
        print('Other channels reader threads finished')

    if COLLECT_STATISTICS:
        print('Clothing statstics channel')
        stat_chl = clt1.srv_rpc.stop_statistic_channel(clt1.client_id)
        time.sleep(2)
        receive_stop.set()
        read_statistics.join()  # This is the last

    print('Last thread finished')

    # 10. close the clients:
    time.sleep(2) #just to be sure we wait a bit longer
    print('Delete the client objects and close the app')
    clt1.stop_client()
    clt2.stop_client()
    clt3.stop_client()
    clt4.stop_client()

    #input('PRESS return to stop') # enable for debbuging proposes



"""
This code is taken from PyPi stream_service package
https://github.com/BR1py/stream_service
"""

from __future__ import absolute_import
import asyncio
import copy
from inspect import signature
from .helpers import *
from .frame import Frame,FrameHeader,CT_DECODE,CT_EXCEPTION

DEBUG=False

# RPC related classes

class RPCMethod():

    def __init__(self,name,signature,header,reader,writer,loop,get_transaction_method=None):
        self._name=name
        self._signature=signature
        self._header=header
        self._reader=reader
        self._writer=writer
        self._loop=loop
        self._get_transaction_method=get_transaction_method

    def __call__(self,*args,**kwargs):
        """
        Here we do the remote call!
        :param args: rpc argumnets
        :param kwargs: rpc keyword arguments
        :return:
        """
        task=asyncio.run_coroutine_threadsafe(self.exec_rpc(*args,**kwargs),self._loop)
        return task.result()

    async def exec_rpc(self,*args,**kwargs):
            """
                Here we do the remote call!
                :param args: rpc arguments
                :param kwargs: rpc keyword arguments
                :return:
                """
            # we check if the parameters match
            self._signature.bind(*args, **kwargs)
            # create transaction
            ta = self._get_transaction_method()
            if DEBUG: print('RPC call transaction %i created'%ta.id)
            # check if tto parameter is given
            if ta.tto_para_name in kwargs:
                tto = kwargs.pop(ta.tto_para_name)
            # create header
            tx_header = copy.copy(self._header)
            tx_header.set_transaction_id(ta.id)
            # create frame and send out for remote execution
            tx_frame=Frame(tx_header, *args, **kwargs)
            await send_msg(self._writer,tx_frame.encode())
            if DEBUG: print('RPC call send to server:',self.get_method_string(args,kwargs))
            rx_bytes = await ta.buffer.get()
            # wait for response in case of timeout a timeout exception will be raised
            back = Frame(FrameHeader(CT_DECODE, address_size=tx_header.address_size), rx_bytes)
            if back is None:
                raise StreamError_Command('No valid return received')
            if back.header.container_type == CT_EXCEPTION:
                raise StreamError_Command(repr(back.args))
            if DEBUG: print('RPC call return:', repr(back.args))
            # for returns we extract just the first item of the args tuple
            args=back.args[0]
            return args

    def get_method_string(self,args,kwargs):
        return '%s(%s, %s)'%(self._name,repr(args),repr(kwargs))

class RPCMethodDescription():
    __slots__=('name','method_obj','argspec','method_id','method_id_bytes','address_size')
    def __init__(self,name,method_idx,argspec,address_size=2):
        '''
        Ths objects holds the whole description of an rpc method
        :param name: method name
        :param method_idx: method idx (used in frame headers)
        :param argspec: argument specification (signature)
        :param address_size: address_size used on the server
        '''
        self.name=name
        self.argspec=argspec
        self.method_id=method_idx
        self.method_id_bytes=method_idx.to_bytes(address_size,BO)
        self.address_size=address_size

    def change_name(self,name):
        self.name=name

    def change_method_idx(self,method_idx):
        self.method_id = method_idx
        self.method_id_bytes = method_idx.to_bytes(self.address_size, BO)

    def get_desc_dict(self):
        return {'name':self.name,
                'argspec': self.argspec}

    def __repr__(self):
        out=['RPCMethodDescription(',
             ' name = %s,'%repr(self.name),
             ' method_idx = %s,' % repr(self.method_id),
             ' argspec = %s)' % repr(self.argspec)
            ]
        return ''.join(out)

class RPCEmpty():
    # class is used to instance the sub classes
    pass
class RPCFactory():
    def __init__(self,rpc_methods,header,reader,writer,loop,get_transaction_method):
        self._rpc_methods=rpc_methods
        self._header=header
        self._reader=reader
        self._writer=writer
        self._loop=loop
        self._get_transaction_method=get_transaction_method
        self._id_to_method = {}
        self._populate_methods()

    def _populate_methods(self):
        '''
        helper function that populate the class with the RPCMethod frames
        '''
        for m_id,method_item in enumerate(self._rpc_methods):
            header=copy.copy(self._header)
            header.set_command_id(m_id)
            m_name=method_item.name
            classes=m_name.split('.')
            parent = self
            if len(classes)>1:
                for pre_name in classes[:-1]:
                    if hasattr(parent,pre_name):
                        parent=parent.__getattribute__(pre_name)
                        if callable(parent):
                            # something is wrong we expect a class here!
                            # we ignore the entry
                            continue
                    else:
                        #create a new empty calss to append the methods
                        parent.__setattr__(pre_name,RPCEmpty())
                        parent=parent.__getattribute__(pre_name)
                m_name=classes[-1]
            rpc_m=RPCMethod(m_name,
                           method_item.argspec,
                           header,
                           self._reader,
                           self._writer,
                           self._loop,
                           self._get_transaction_method)
            parent.__setattr__(m_name,rpc_m)
            self._id_to_method[m_id]=rpc_m

    @staticmethod
    def class_init_rpc_functions(class_object,first_items=[],address_size=2):
        '''
        helper function which analysis the available rpc methods and creates the related description
        e.g. used during server or client init
        :param class_object: The main class object that should be analysed (server or client)
        :return:
        '''

        attributes = list(dir(class_object))
        #reorder first items:
        i=0
        for item in first_items:
            if item in attributes:
                attributes.remove(item)
                attributes.insert(i,item)
                i+=1

        rpc_methods_descs=[]
        rpc_methods=[]

        for attribute in attributes:
            try:
                if attribute.startswith('rpc_'):
                    if callable(getattr(class_object, attribute)):
                        # rpc method
                        idx = len(rpc_methods_descs)
                        data = None
                        m = class_object.__getattribute__(attribute)
                        try:
                            data = signature(m)
                        except:
                            pass
                        md = RPCMethodDescription(attribute[4:], idx, data, address_size=address_size)
                        rpc_methods.append(m)
                        rpc_methods_descs.append(md)
                    else:
                        # rpc class analyse the class for rpc commands and append to the description
                        rpc_methods_desc2,rpc_methods2=RPCFactory.class_init_rpc_functions(getattr(class_object, attribute), address_size=address_size)
                        rpc_sub_class=RPCEmpty()
                        for md,m in zip(rpc_methods_desc2,rpc_methods2):
                            rpc_sub_class.__setattr__(md.name,m)
                            md.change_name(attribute[4:]+'.'+md.name)
                            md.change_method_idx(len(rpc_methods_descs))
                            rpc_methods_descs.append(md)
                            rpc_methods.append(m)
            except:
                # we ignore methods which raises exceptions
                pass
        return rpc_methods_descs,rpc_methods




"""
This code is taken from PyPi stream_service package
https://github.com/BR1py/stream_service
"""
import asyncio
import random

#DEFAULT SERVER PORT
DEFAULT_BUFFER_STREAM_MGR_PORT = 50012

#Possible buffer full reactions on server
# Run the buffer queue as a ringbuffer (delete oldest item and put new in)
RING_BUFFER_FULL=0 # Default behavior
# Run the buffer queue in clear mode which means if full delete all items and put new item in
CLEAR_BUFFER_FULL=1
# Run the buffer queue in skip mode which means if full the new item will be ignored and the old items are kept in the buffer queue
SKIP_BUFFER_FULL=2

#Stream Server/Client related Exceptions:

class StreamError_Authenticate(Exception):
    pass
class StreamError_Command(Exception):
    pass
class StreamError_Overload(Exception):
    pass
class StreamError_Connection(Exception):
    pass
class StreamError_RPC(Exception):
    pass
class StreamError_Channel(Exception):
    pass
class StreamError_Target(Exception):
    pass
class StreamError_Coding(Exception):
    pass
class StreamError_TransactionMissmatch(Exception):
    pass

BO='big' # byte order used in the frames

TTO_DEFAULT = 20  # Default RPC timeout (can be reset via StreamClient parameter)
TTO_DEFAULT_PARA_NAME = 'tto'  # Default parameter name for RPC timeout (can be reset via StreamClient parameter)


class Logger():
    MAX_LOG_CHAR = 10000

    def __init__(self, log_file, header=None, start_tag='', quiet=False):
        self.start_tag = start_tag
        self.char_nr = 0
        self.log_file = log_file
        self.quiet = quiet
        if log_file is not None:
            if os.path.exists(log_file):
                self.create_bak(quiet=True)
            if header is not None:
                self.write_log(header)

    def create_bak(self, quiet=False):
        log_file = self.log_file
        if log_file is not None:
            if os.path.exists(log_file):
                if os.path.exists(log_file + '.bak'):
                    os.remove(log_file + '.bak')
                os.rename(log_file, log_file + '.bak')

    def write_log(self, value):
        if self.char_nr > self.MAX_LOG_CHAR:
            self.create_bak()
            self.char_nr = 0
        if not self.quiet:
            print('%s%s' % (self.start_tag, str(value)))
        if self.log_file is not None:
            with open(self.log_file, 'a') as fh:
                fh.write('%f %s%s\n' % (time.time(), self.start_tag, value))
        self.char_nr += len(value)

async def read_msg(stream=asyncio.StreamReader):
    try:
        size_bytes = await stream.readexactly(4)
        size = int.from_bytes(size_bytes, byteorder=BO)
        data = await stream.readexactly(size)
    except:
        return None
    return data

async def send_msg(stream=asyncio.StreamWriter, data=b''):
    size_bytes = len(data).to_bytes(4, byteorder=BO)
    stream.writelines([size_bytes,data])
    await stream.drain()

def int_address_to_bytes(address_int, address_size,mask=None):
    if mask is None:
        mask = 0
        for i in range(address_size):
            mask = mask<<1 + 0xFF
    new_int = address_int & mask
    return new_int.to_bytes(address_size, BO)

def get_args(*args, **kwargs):
    return (locals().values())


def get_raw_key(seed, key):
        '''
        method hides the key in the seed
        (no real encryption but unusual)

        :param seed: seed bytes
        :param key: seed key bytes
        :return: raw_key to be compared
        '''
        # in first step the bytes are mixed
        new = b''
        kl = len(key)
        i = 0
        for i in range(len(seed)):
            if i < kl:
                new = key[i].to_bytes(1, BO) + new
            new = seed[i].to_bytes(1, BO) + new
        if i < kl:
            new = key[i:] + new
        new_int = int.from_bytes(new[:16], BO)
        key_int = int.from_bytes(key, BO)
        raw_int = hash((new_int * key_int))
        if raw_int < 0:
            raw_int = raw_int * -1
        raw_key = (raw_int).to_bytes(8, BO)
        return raw_key


class IdsHandler():

    def __init__(self, address_size):
        """
        This is a helper class for handling (generating and management) of ids
        in the given address size
        The methods deliver ids from a given set and put them back into the set in case they are no longer used
        :param address_size: address size to use for the generation of the ids
        """
        mask = 0
        for i in range(address_size):
            mask = (mask << 8) + 0xFF
        # prepare random generator
        self._choices = {i for i in range(mask)}
        self._random = random.Random()
        self._random.seed()

    def get_new_id(self):
        new_id =  self._random.sample(self._choices,k=1)[0]
        self._choices.remove(new_id)
        return new_id

    def free_id(self, old_id):
        self._choices.add(old_id)



"""
This code is taken from PyPi stream_service package
https://github.com/BR1py/stream_service

Stream Buffer Server main module

The naming in this module is always from the perspective of the server!
client writer -> transport data to the client
client reader -> read data from the client

The address_size can be defined in the server. It gives the number of bytes used for all kinds of addressing:
server_ids
channel_ids
client_ids
rpc_command_ids
data_package counter
transaction_ids
...

The packages exchanged have the following form

1. server_address for mirror servers (=0 if no server function should be mirrored)
2. channel_id (=0 cannot be used directly this is the server channel which is used for channel control, server control and rpc calls)
3. type byte: byte contains container type and code information
4. data package counter
5. rest of the package is the payload
# if container type is command
4. command id
5. rest are the command arguments

About RPC calls

the rpc calls are handled via server channel 0
Each client can call a exposed rpc function of the other client here.
All calls are asynchronous which means if the call request is send a transaction id is returned.
After the call is calculated the return is given with the related transaction id to the client.

The client object itself contains async or snyc calls handling for sync calls the Client will block the exeution and
wait for the return with the right transaction id.

"""
from __future__ import absolute_import

import asyncio
import traceback
from threading import Lock
from multiprocessing import Process
from asyncio import StreamReader, StreamWriter, QueueFull
import uuid
import time
import threading
from .lib import *
from .lib.frame import *
from .lib.buffer import *

# from .stream_commands import *
DEBUG = False  # Variable used for development proposes (more print outputs given)
               # curious leads into an exception when set -> I must investigate

__package__ = 'stream_service'
__version__ = '0.1.0'
__licence__ = 'MIT'
__author__ = 'B.R.'
__url__ = 'https://github.com/BR1py/stream_services'
__description__ = 'Client/Server implementation for data streaming in channels based on python >= 3.5'


# helper classes for the server:


class ServeClient():
    __slots__ = ('_peername', '_id', '_name', '_control_key', '_async_reader', '_async_writer', '_rpc_methods_descs',
                 '_read_task', '_write_task', '_owned_chl_ids', '_write_chls', '_read_bufs', '_alive', '_rpc_methods',
                 '_send_cnt', '_id_bytes')

    def __init__(self, peername, client_id, id_bytes, name, control_key, async_reader, async_writer, rpc_methods_descs):
        """
        Object contains all client related information for each client that is connected to the sever

        :param peername: Adresse of the client host
        :param client_id: unique id of the client
        :param id_bytes: client id already converted to bytes
        :param name: name of the client
        :param control_key: authentication key to delete the client
        :param async_reader: asyncio StreamReader of the client (communication channel from the client to the server
        :param async_writer: asyncio StreamWriter of the client (communication channel to the client from the server
        :param rpc_methods_descs: rpc method description of the client (to be shared with other clients)
        """
        self._peername = peername
        self._id = client_id
        self._name = name
        self._control_key = control_key
        self._id_bytes = id_bytes
        self._async_reader = async_reader
        self._async_writer = async_writer
        self._rpc_methods_descs = rpc_methods_descs

        self._read_task = None # here we store the client read task
        self._owned_chl_ids = set() # ServeChannel object owned by this client
        self._write_chls = {} # ServeChannel writers subscribed by the client
        self._read_bufs = set() # ReadBuffer for the ServeChannels the client subscribed for reading
        self._alive = True # alive flag for the _read_task
        self._send_cnt = 0 # counter for statistics for the send frames

    # make some private elements available
    @property
    def id(self):
        return self._id

    @property
    def id_bytes(self):
        return self._id_bytes

    @property
    def name(self):
        return self._name

    @property
    def is_alive(self):
        return self._alive

    @property
    def writer(self):
        return self._async_writer

    @property
    def control_key(self):
        return self._control_key

    def stop(self):
        """
        stop the read_task by switching the alive flag to False
        (The task will stop after timeout of the read function is reached or when a read item is received)
        """
        self._alive = False

    def register_owned_chl(self, chl):
        """
        Helper to register a new owned channel (used in the rpc_create_new_channel() method of the server)
        (An owned channel is automatically appended as a writer channel)
        :param chl: ServeChannel object that is owned by the channel
        """
        chl_id = chl.id
        self._owned_chl_ids.add(chl_id)
        self._write_chls[chl_id] = chl
        self._write_chls[chl.name] = chl

    def set_read_task(self, task):
        """
        Helper method to set the read task (used in the _connect() method of teh server)
        :param task: give the asyncio task with the writer
        :return:
        """
        self._read_task = task

    def set_send_cnt(self, cnt):
        """
        helper to set the actual send count (used by the reader task)
        :param cnt: new counter value
        """
        self._send_cnt = cnt

    def get_async_stream_objects(self):
        """
        Helper to get the clients asyncio StreamReader and StreamWriter of the client
        :return: StreamReader,StreamWriter
        """
        return self._async_reader, self._async_writer

    def get_rpc_method_desc(self):
        """
        get the rpc method description of the client
        :return:
        """
        return self._rpc_methods_descs

class ServeChannel():
    __slots__ = ('_parent', '_id', '_name', '_owner', '_is_public', '_is_single_in', '_lifetime',
                 '_read_bufs', '_control_key', '_transaction_ids', '_address_mask', '_read_writers',
                 '_channel_task', '_task_trigger', '_task_alive', '_loop', '_slot_time', '_write_cnt',
                 '_next_send_time', '_is_slot_time_fixed', '_address_size', '_statistic_chl', '_event_cnt',
                 '_statistic_chl_fill_rate', '_writer_clients', '_ta_ids_hdl')

    def __init__(self, chl_id, name, owner, is_public, is_single_in, lifetime, loop, address_size,
                 fixed_slot_time=None):
        """
        This class represents a channel on the server and it contains an own task to send the data to the subscribers
        It's the most important class in the server!

        :param chl_id: channel id
        :param name: channel name
        :param owner: channel owner (client that has created the class (this client must also delete the channel))
        :param is_public: Will the channel be published by the server (server gives an info list if requested)
        :param is_single_in: Only one writer is allowed for this channel
        :param lifetime: temporary channels can be created too they will be eliminated after lifetime (in seconds)
        :param loop: event loop of the server
        :param address_size: address size used on this server
        :param fixed_slot_time: channel has a fixed slot time (load balancing disabled for this channel)
        """
        self._id = chl_id
        self._name = name
        self._owner = owner
        self._is_public = is_public
        self._is_single_in = is_single_in
        self._lifetime = lifetime
        self._address_size = address_size
        # Create a mask based on address_size:
        mask = 0
        for i in range(address_size):
            mask = (mask << 8) + 0xFF
        self._address_mask = mask
        self._control_key = uuid.uuid4().bytes # need for channel delete
        self._ta_ids_hdl = IdsHandler(address_size) # Generater for Transaction ids used on this channel
        self._read_bufs = {} # subscribed read buffers
        self._writer_clients = {} # subscribed writers
        self._task_alive = True  # alive flag for the channel read task will be set to false to stop the task
        self._loop = loop # event loop that should be used (the one from the sevrer)
        self._task_trigger = asyncio.Event(loop=loop) # activity trigger for the read task used by client reader tasks to trigger the channel writer task
        self._channel_task = self._loop.create_task(self._write_task()) # writer task is started here
        if fixed_slot_time is None: # a fixed slot time can be set for a channel if not the slottime is set depending of the number of channels on the server
            self._slot_time = 2  # default is 2 seconds
            self._is_slot_time_fixed = False
        else:
            self._slot_time = fixed_slot_time  # default is 2 seconds
            self._is_slot_time_fixed = True
        self._next_send_time = 0 # timing used for load balancing
        self._write_cnt = 0 # statistics counter
        self._event_cnt = 0 # statistics counter
        self._statistic_chl = None # statistics channel (if set statistic data from the channel is put in this channel
        self._statistic_chl_fill_rate = 1 # helper for statistic channel (gives the fill rate for this channel

    # publish some internals

    @property
    def is_slot_time_fixed(self):
        return self._is_slot_time_fixed

    @property
    def write_cnt(self):
        return self._write_cnt

    @property
    def last_send_time(self):
        return (self._next_send_time - self._slot_time)

    @property
    def control_key(self):
        return self._control_key

    @property
    def name(self):
        return self._name

    @property
    def id(self):
        return self._id

    @property
    def is_single_in(self):
        return self._is_single_in


    def add_client_writer(self, client):
        """
        add a new writer to the channel
        :param client: writer client
        :return:
        """
        self._writer_clients[client.id] = client

    def remove_client_writer(self, client_id):
        """
        remove a writer from the channel
        :param client_id: client id that should be removed
        :return:
        """
        if client_id in self._writer_clients:
            del self._writer_clients[client_id]
            return True
        return False

    def add_client_reader(self, buffer):
        '''
        add a given read buffer to the channel
        :param buffer: Buffer() object (we expect that the additional kwargs in the object a set as expected ->
                       this is normally ensured in the rpc method of the server)
        :return:
        '''
        client_id = buffer.client.id
        if client_id in self._read_bufs:
            raise StreamError_Channel(
                'Given client_id: %s exists already as reader in the channel, delete first!' % repr(client_id))
        if len(self._read_bufs) == 0:
            self._notify_active(True)
        self._read_bufs[client_id] = buffer

    def remove_client_reader(self, client_id):
        """
        remove a client reader from the channel
        :param client_id:
        :return:
        """
        if client_id in self._read_bufs:
            raise StreamError_Channel(
                'Given client_id: %s not found in channel readers, cannot delete!' % repr(client_id))
        read_buffer = self._read_bufs.pop(client_id)
        # we clear the buffer and we send as last item an end data object
        read_buffer.clear()
        header = FrameHeader(CT_DATA_END, CODE_MARSHAL, address_size=self._address_size)
        read_buffer.put_nowait(header, b'END DATA')
        if len(self._read_bufs) == 0:
            self._notify_active(False)
        return read_buffer

    def _notify_active(self, active=True):
        """
        signalize if put_data activity is needed for this channel
        To avoid useless traffic on the network the writers for this channel can be deactivated
        in case no subscriber listens to the channel
        :param active: True - activity on, False activity of
        :return:
        """
        for client in self._writer_clients.values():
            frame = Frame(FrameHeader(CT_CMD, CODE_MARSHAL, command_id=1,
                                      # rpc_command for switch the activity on clients must have id==1 !
                                      client_id=client.id), self.id, active)
            task = asyncio.run_coroutine_threadsafe(send_msg(client._async_writer, frame.encode()), self._loop)
            task.result()

    def set_slot_time(self, slot_time):
        """
        slot time can be adapted in case of server load or to prioritize a channel
        :param slot_time: slot time for sending
        :return:
        """
        if not self._is_slot_time_fixed:
            self._slot_time = slot_time

    def set_statistic_channel(self, chl, fill_rate):
        """
        The statistic channel can be activated on the server to track the load,
        with this method the channel is set and activated for this channel
        :param chl: statistics channel
        :param fill_rate: fill_rate for the statistic channel
        :return:
        """
        self._statistic_chl = chl
        self._statistic_chl_fill_rate = fill_rate

    def stop_channel(self):
        """
        Will stop the channel read task (cannot be restarted!)

        :return:
        """
        self._task_alive = False
        self._task_trigger.set()

    async def _write_task(self):
        """
        This tasks ensures that the data is send only in the given timeslot
        Therefore we do not send directly we buffer in queues
        """
        # prepare some locals for quicker access
        # (The task should run as quick as possible to reduce the load on the server)
        read_buffers = self._read_bufs
        buf_infos = []
        statistics_channel_id = None
        statistic_header = None
        statistics_channel_fill_rate = 1
        last_statistics_event_cnt = -1
        while self._task_alive: # Main write loop
            # We wait for a trigger event signalize that we have new data
            await self._task_trigger.wait()
            # clear the event to be usable again
            self._task_trigger.clear()
            # print('Trigger received')
            await asyncio.sleep(0.001)  # give other tasks a slot
            if self._statistic_chl is not None:
                # if statistics channel is active we measure some buffer statistics before we put items out
                buf_infos = [(buf.client.name, buf.qsize(), buf.full_cnt, buf.fill_cnt) for k, buf in
                             self._read_bufs.items() if
                             type(k) is int]
                statistics_active = True
            else:
                statistics_active = False
            # init some variables for break conditions
            cnt = 0
            t = self._loop.time()
            end_time = t + self._slot_time
            # take data from the buffers and send them to the clients:
            while self._task_alive:
                all_empty = True
                for buffer in list(read_buffers.values()):
                    if not buffer.empty():
                        await asyncio.sleep(0.001)  # give other tasks a slot
                        tx_data = await buffer.get() # take buffer data out (we might use here get_nowait too but hen we must handle the exception)
                        buffer.task_done()  # we free the queue for next operation
                        await asyncio.sleep(0.001)  # give other tasks a slot
                        await self._send_encoded(buffer.client._async_writer, tx_data) # send data to the client
                        # print('Send to client',tx_data)
                        cnt = (cnt + 1) & 0b11111111
                        all_empty = False
                # check break conditions:
                # we break earliest after sending at least one item per read_buffer
                if end_time > self._loop.time():
                    # time slot over
                    self._write_cnt = cnt
                    self._next_send_time = self._loop.time() + self._slot_time
                    await asyncio.sleep(0.001)
                    # set trigger (we are not finished)
                    self._task_trigger.set()
                    break
                if all_empty:
                    # all read buffers are empty -> Fine no data lost in this case!
                    self._write_cnt = cnt
                    self._next_send_time = self._loop.time() + self._slot_time
                    await asyncio.sleep(0.1)  # here we wait a bit longer (the buffers are well cleaned)
                    break
            if statistics_active is not None:
                # if statistics channel is active we send statistics data
                # this will slow down the task!
                if not last_statistics_event_cnt == cnt:
                    try:
                        # We put this in try except because in a corner-case the statistic channel might be set to None
                        # during this operation!
                        chl_id = self._statistic_chl.id
                        if statistics_channel_id != chl_id:
                            # channel changed!
                            # create new header
                            statistic_header = FrameHeader(CT_DATA, CODE_MARSHAL, channel_id=chl_id)
                            statistics_channel_id = chl_id
                            if statistics_channel_id == self.id:
                                # we do not use a fill_rate on the statistics channel itself
                                statistics_channel_fill_rate = 1
                            else:
                                statistics_channel_fill_rate = self._statistic_chl_fill_rate
                        if not cnt % statistics_channel_fill_rate:
                            info = '%f %s(%s): send_count: %i; event_count: %i; buffer sizes/full_cnt/fill_cnt: %s\n' % (
                                t, self._name, self._id, cnt, self._event_cnt, repr(buf_infos))
                            frame = Frame(statistic_header, info)
                            self._statistic_chl.put_nowait(CT_DATA, frame.encode())
                            last_statistics_event_cnt = cnt
                    except:
                        pass


    def get_new_transaction_id(self):
        """
        We do not have transaction objects on server like on the clients but we use transaction ids in some cases
        Here we can create unique ids
        Will raise an Exception IndexError (or ValueError) in case no more id is available
        :return:
        """
        return self._ta_ids_hdl.get_new_id()

    def clear_buffers(self):
        """
        clear all the buffers
        :return:
        """
        for read_buffer in list(self._read_bufs.values()):
            read_buffer.clear()

    def put_nowait(self, container_type, data):
        """
        put data function used by the client writers
        Inside we distribute the data to the buffers
        :param container_type: This info is already parsed by the reader thread and we reuse it here
        :param data: raw data received for this channel
        :return:
        """
        force = False
        if container_type == CT_DATA_END:
            # DATA :END RECEIVED!
            # In case an data end is received the transaction id must be freed for reuse
            header = FrameHeader(CT_DECODE, data, self._address_size)
            self._ta_ids_hdl.free_id(header.transaction_id)
            force=True
        self._event_cnt = (self._event_cnt + 1) & 0xFFFFFFFF #statistics counter
        for read_buf in list(self._read_bufs.values()): # put data into the buffers
            read_buf.put_nowait(data,force=force)
        if self._loop.time() > self._next_send_time:
            # we set the task trigger event to signalize the send task that we have new data to be send
            self._task_trigger.set()

    async def put(self, container_type, data):
        """
        async version of put_nowait but this is not used at the Moment
        :param container_type:
        :param data:
        :return:
        """
        if container_type == CT_DATA_END:
            # DATA :END RECEIVED!
            header = FrameHeader(CT_DECODE, data, self._address_size)
            self._ta_ids_hdl.free_id(header.transaction_id)
        for read_buf in list(self._read_bufs.values()):
            await read_buf.put(data)
        if self._loop.time() > self._next_send_time:
            self._task_trigger.set()

    async def _send_encoded(self, writer, tx_data):
        '''
        helper method to encode the send data
        the encoding is done depending on the tx_data type given

        :param writer: writer
        :param tx_data: tx_data can be bytes or tuple header and payload or a Frame() object
        :return:
        '''
        if type(tx_data) is tuple:  # we have header object and payload!
            await send_msg(writer, tx_data[0].encode() + tx_data[1])
        elif type(tx_data) is bytes:  # we have raw bytes
            await send_msg(writer, tx_data)
        else:  # we have a tx_frame
            await send_msg(writer, tx_data.encode())


# Server:

class StreamChannelServer_Process(Process):

    def __init__(self, ip_address=None, port=None, ssl=None,
                 address_size=2, log_file=None, quiet=False, name='StreamDataServer',
                 client_authenticate=b'STRM_Client',
                 # scale and performance parameters
                 max_clients=100,
                 max_slottime=30,
                 min_slottime=2,
                 default_limit=2 ** 16,
                 statistic_channel_name='ServerChannelStatistics',
                 daemon=False, *kwargs):
        """
        Server Process - main stream service server class
        Important parts of the class init are executed in the moment the process is started (see run() method)
        :param ip_address: network IP address that will be used by the server
        :param port: network port that will be used by the server
        :param ssl: optional ssl configuration (search for  sll on asyncio servers for more details)
        :param address_size: address_size used on the server default is 2 bytes
                             e.g. if you need more channels this can be extended but the frame size increases in this case!
        :param log_file: USe a log file to store the log messages of the server (the file will be limited in size)
        :param quiet: run in quiet mode (no log outputs printed into the terminal)
        :param name: name of the server
        :param client_authenticate: authentication bytes for clients (need for connections)
        :param max_clients: maximum number of clients accepted by the server
        :param max_slottime: maximum slot time might be increased depending on the machine the server is running on;
                             Used for load balancing, it's the upper limit for the slot-time a channel has to send data
                             and it is divided by the number of channels on the server
                             (normally the parameter should be kept as it is)
        :param min_slottime: minimum slot time might be increased depending on the machine the server is running on
                             Used for load balancing; if a lot of channels running on the server the slot-time
                             for writing is fixed at this value
                             (normally the parameter should be kept as it is
        :param default_limit: This is the size of the StreamWriter buffer (normally this should not be changed)
        :param statistic_channel_name: name of the statistic channel
        :param daemon: run the process in daemon mode
        :param kwargs: additional arguments to give to the asyncio server
        """
        if port is None:
            port = DEFAULT_BUFFER_STREAM_MGR_PORT
        self._address = (ip_address, port)
        self._kwargs=kwargs
        self._address_size = address_size
        self._max_slottime = max_slottime
        self._min_slottime = min_slottime
        self._default_limit = default_limit
        self._current_max_writer_size = self._default_limit
        self._current_slot_delta_time = self._max_slottime
        self._ssl = ssl
        self._statistic_channel_name = statistic_channel_name
        self._statistic_chl = None
        self._statistic_chl_fill_rate = 1
        mask = 0
        for i in range(address_size):
            mask = (mask << 8) + 0xFF
        self._address_mask = mask
        # prepare random generator
        super().__init__(name=name, daemon=daemon)
        self._srv_lock = None  # lock must be instanced in run
        self._server_ready = False
        self._server_active = True
        self._logger = Logger(log_file=log_file, quiet=quiet)
        self._write_log = self._logger.write_log
        if type(client_authenticate) is not bytes:
            raise TypeError('Given client_authenticate parameter must be of type bytes')

        self._client_authenticate = client_authenticate
        self._clients = {}
        self._max_clients = max_clients
        self._channels = {}
        # the following two items are not used in the channels only on the server for client rpc calls
        self._server_transactions={}
        self._server_transaction_id_hdl = IdsHandler(address_size)

        # id handlers
        self._client_id_hdl = IdsHandler(address_size)
        self._channel_id_hdl = IdsHandler(address_size)

        self._rpc_methods_descs = []
        self._rpc_methods = []

    @property
    def address_size(self):
        return self._address_size

    def run(self):
        try:
            # post initialization:
            self._write_log('Start StreamChannelServer (Version %s)' % __version__)
            self._srv_lock = Lock()
            self._rpc_methods_descs, self._rpc_methods = rpc.RPCFactory.class_init_rpc_functions(self, first_items=[
                'rpc_authenticate_client'],address_size=self._address_size)
            #print(self._rpc_methods_descs)
            #Event loop
            self._loop = loop = asyncio.get_event_loop()
            #self._loop.set_debug(DEBUG)
            #prepare the asyncio server:
            co_rtn = asyncio.start_server(self.client_connect, host=self._address[0], port=self._address[1], loop=loop,
                                          limit=self._default_limit, ssl=self._ssl,**self._kwargs)
            self._server = loop.run_until_complete(co_rtn) # this will be executed in the moment the event loop is started!
            # Serve requests until Ctrl+C is pressed
            self._write_log('Serving on {}'.format(self._server.sockets[0].getsockname()))
            self._server_ready = True
            # start the event loop and run for ever!
            loop.run_forever()
        except:
            self._write_log('Exception in StreamChannelServer:\n')
            self._write_log(''.join(traceback.format_exc()))

    async def client_connect(self, reader: StreamReader, writer: StreamWriter):
        '''
        request from a client for a new connection
        this is the asyncio standard connect functionality
        The return of this method is send to the client
        Note:: In this method the internal client related objects are created especially the two tasks
        for reading and sending from/to the client
        State: code tested
        :param reader: client StreamReader object
        :param writer: client StreamWriter object
        '''
        rx_header = None
        address_size = self._address_size
        try:
            # get the peername
            peername = writer.get_extra_info('peername')
            self._write_log('client_connect(): Client Request from %s' % str(peername))
            # is the server full?
            # we must divide by 2 because we store each client two times in the dict (id and name)!
            if len(self._clients) / 2 >= self._max_clients:
                self._write_log('client_connect(): Client Request denied, to many clients connected')
                # negative response server full
                await send_msg(writer, b'0')
                return
            # Server starts communication and send address_size and seed
            # here we put some small security in
            # for more one must use ssl!
            seed = uuid.uuid4().bytes + uuid.uuid4().bytes
            raw_key = get_raw_key(seed, self._client_authenticate)
            if DEBUG: print('client_connect(): seed+key pair is:', seed, raw_key)
            tx_bytes = address_size.to_bytes(1, BO) + seed
            if DEBUG: print('client_connect(): First message send is:', tx_bytes)
            await send_msg(writer, tx_bytes)
            # waiting for clients response with authentication info
            rx_bytes = await read_msg(reader)
            if DEBUG: print('client_connect(): Package raw bytes received:', rx_bytes)
            # decode received answer
            rx_frame = Frame(FrameHeader(CT_DECODE, address_size=address_size), rx_bytes)
            rx_header = rx_frame.header
            if DEBUG: print('client_connect(): Frame received:', repr(rx_frame))
            # we already prepare the response header here, will be used in the exceptions too!
            tx_header = FrameHeader(container_type=CT_RETURN,
                                    coder_type=CODE_PICKLE,
                                    transaction_id=rx_header.transaction_id,
                                    address_size=address_size)
            if rx_header.container_type != CT_SRV_CMD and rx_header.command_id != 0:
                self._write_log('client_connect(): Client Request denied, wrong connect command send')
                raise StreamError_Command('client_connect(): Wrong connect command send')
            # authenticate client
            args = rx_frame.args
            back = self.rpc_authenticate_client(*args, raw_key=raw_key)
            if back[0] != 0:
                self._write_log('client_connect(): Client Request denied, authentication failed')
                raise StreamError_Authenticate('client_connect(): Client Request denied, authentication failed')
            # extract other data
            client_name = back[1]
            rpc_methods_descs = back[2]
            if DEBUG: print('client_connect(): Client authentication finished')
            # create a new client object
            client_id = self._client_id_hdl.get_new_id()
            if client_id is None:
                self._write_log('client_connect(): Client Request denied, to many clients connected')
                raise StreamError_Authenticate('client_connect(): Client Request denied, no free id found')
            control_key = uuid.uuid4().bytes
            id_bytes = client_id.to_bytes(address_size, BO)
            client = ServeClient(peername, client_id, id_bytes, client_name,
                                 control_key, reader, writer, rpc_methods_descs)
            # enter client in the client dict
            with self._srv_lock:
                # we add the client two times in the client dict because we like to reach it quick per name or id
                self._clients[client_id] = client
                self._clients[client_name] = client
                client.set_read_task(self._loop.create_task(self._client_task_loop(client)))
                # rescale the writer limits:
                l = len(self._clients) / 2
                self._current_max_writer_size = self._default_limit / l
                self._current_slot_delta_time = self._max_slottime / l
                if self._current_slot_delta_time < self._min_slottime:
                    self._current_slot_delta_time = self._min_slottime
                # asyncio.gather(client.read_task,client.read_task)
            if DEBUG: print('client_connect(): Server related Client object created', client)
            # Create the Client default channel
            # create positive response
            tx_frame = Frame(tx_header, client_id, control_key, self._rpc_methods_descs)
            if DEBUG: print('client_connect(): Positive response pkg build', tx_frame)
            tx_bytes = tx_frame.encode()
            if DEBUG: print('client_connect(): Positive response send', tx_bytes)
            # time.sleep(2)  # give the loops some time to come up
            # give the feedback that client setup is finished
            await send_msg(writer, tx_bytes)
            self._write_log('client_connect(): Client connection established all server internal objects created')
        except Exception as e:
            # Catch exception and if possible send back to the client
            self._write_log('client_connect(): Exception:')
            self._write_log(repr(e))
            self._write_log(traceback.format_exc())
            if rx_header is None:
                # give unspecific feedback
                try:
                    await send_msg(writer,
                                   Frame(FrameHeader(CT_EXCEPTION, CODE_MARSHAL),
                                         repr(e),
                                         traceback.format_exc()).encode())
                except:
                    self._write_log(
                        'client_connect(): Unspecific Exception return not send to client; sub-exception traceback: '
                        '%s' % ''.join(traceback.format_exc()))
            else:
                # give specific feedback
                try:
                    await send_msg(writer,
                                   Frame(FrameHeader(CT_EXCEPTION, CODE_MARSHAL,
                                                     transaction_id=rx_header.transaction_id,
                                                     command_id=rx_header.command_id,
                                                     channel_id=rx_header.channel_id,
                                                     client_id=rx_header.client_id,
                                                     counter=rx_header.counter),
                                         repr(e), ''.join(traceback.format_exc())).encode())
                except:
                    self._write_log('client_connect(): Exception return not send to client; sub-exception traceback: '
                                    '%s' % ''.join(traceback.format_exc()))

    async def _send_encoded(self, writer, tx_data):
        '''
        helper method to encode the send data
        the encoding is done depending on the tx_data type given

        :param writer: writer
        :param tx_data: tx_data can be bytes or tuple header and payload or a Frame() object
        :return:
        '''
        if type(tx_data) is tuple:  # we have header object and payload!
            await send_msg(writer, tx_data[0].encode() + tx_data[1])
        elif type(tx_data) is bytes:  # we have raw bytes
            await send_msg(writer, tx_data)
        else:  # we have a tx_frame
            await send_msg(writer, tx_data.encode())

    async def _client_task_loop(self, client):
        # This is the client reader loop
        # prepare some locals for quicker access
        reader, writer = client.get_async_stream_objects()
        address_size = self._address_size
        id_bytes = client.id_bytes
        if client._alive:
            self._write_log('Client %s read from client loop/task started' % client.name)
        none_cnt = 0
        loop_cnt = 0
        first_id_slice = slice(2, self._address_size + 2)
        second_id_slice = slice(2 + self._address_size, 2 * self._address_size + 2)
        while client.is_alive:
            # loop_cnt+=1
            await asyncio.sleep(0.001)  # this allows other threads to be executed in parallel
            rx_header = None
            try:
                rx_bytes = False
                try:
                    rx_bytes = await asyncio.wait_for(read_msg(reader),
                                                      10)  # timeout is used just to clean the client if it no longer exists
                except:
                    rx_bytes = False
                await asyncio.sleep(0.001)  # this allows other threads to be executed in parallel
                # print('RX',rx_bytes)
                if rx_bytes is None:
                    # in case of None the client might be died but we give the loop some more tries
                    none_cnt += 1
                    if none_cnt > 10:
                        # client died!
                        client.stop()
                        self._write_log('Client %s/%s connection error, connection will be killed from server' % (
                            str(client.name), str(client.id)))
                        break
                    continue
                elif rx_bytes is not False:
                    if DEBUG: print('Frame received:', rx_bytes)
                    # we got a valid frame we reset the none counter:
                    none_cnt = 0
                    # Take data from clients and put it in the write queues
                    # extract the container type
                    container_type = rx_bytes[0] & 0b1111
                    # depending of the container type we react
                    if container_type == CT_DATA or container_type == CT_DATA_END:
                        chl_id = int.from_bytes(rx_bytes[first_id_slice], BO)
                        # channel related data we put into the channel send queue
                        chl = self._channels.get(chl_id)
                        if chl is None:
                            raise StreamError_Channel('Given channel_id %s is unknown or not writeable' % (chl_id))
                        await asyncio.sleep(0.001)  # this allows other threads to be executed in parallel
                        try:
                            chl.put_nowait(container_type, rx_bytes)
                        except:
                            print(''.join(traceback.format_exc()))
                        await asyncio.sleep(0.001)  # this allows other threads to be executed in parallel
                        if DEBUG: print('Data frame received and put in channel %i:' % chl_id, rx_bytes)
                    elif container_type == CT_RETURN or container_type == CT_CMD:
                        # here we route from one client to another client
                        if rx_bytes[1] & 0b10 == 0:
                            # no client given -> server request response
                            # we ignore
                            continue
                        target_client_id = int.from_bytes(rx_bytes[first_id_slice], BO)
                        if target_client_id == 0:
                            # this is a return to a server call return we ignore
                            # e.g. from channel activation calls
                            continue
                        if target_client_id is None or target_client_id == client.id:
                            # no client_id or this client_id is given we just send to the client
                            await asyncio.sleep(0.001)  # this allows other threads to be executed in parallel
                            await send_msg(writer, rx_bytes)
                        else:
                            # the return is target to another client
                            # we exchange the client_id with this client to mark the source and send it ot the target
                            target_client = self._clients.get(target_client_id)
                            if target_client is not None:
                                # in following step client_id is changed to source
                                await asyncio.sleep(0.001)  # this allows other threads to be executed in parallel
                                # exchange target client id with source client id:
                                new_bytes = b''.join([rx_bytes[:first_id_slice.start],
                                                      client.id_bytes,
                                                      rx_bytes[first_id_slice.stop:]
                                                      ])
                                await send_msg(target_client._async_writer, new_bytes)
                            else:
                                raise StreamError_Target(
                                    'Unknown target (%s) given, return cannot be send' % (repr(target_client_id)))
                    elif container_type == CT_SRV_CMD:
                        t = threading.Thread(target=self._execute_rpc_threaded, args=(client, rx_bytes))
                        t.start()
                        continue
                        rx_header = FrameHeader(CT_DECODE, rx_bytes, address_size=address_size)
                        rx_payload = rx_bytes[len(rx_header):]
                        # server rpc calls are executed here
                        try:
                            srv_cmd = self._rpc_methods[rx_header.command_id]
                        except IndexError:
                            srv_cmd = None
                        if srv_cmd is None or srv_cmd == self.rpc_authenticate_client:
                            # the authentication command (cmd_id==0) is protected and is used only in the client_connect()
                            raise StreamError_Command('StreamServer - RPC no/unknown command given')
                        # prepare the command arguments
                        rx_frame = Frame(rx_header)
                        rx_frame.decode_payload(rx_payload)
                        # we replace the client argument with the client object
                        if len(rx_frame.args) > 0:
                            args = [client] + list(rx_frame.args[1:])
                        else:
                            args = [client]
                        kwargs = rx_frame.kwargs
                        if 'client_id' in kwargs:
                            del kwargs['client_id']
                        # execute the command
                        back = srv_cmd(*args, **kwargs)
                        # prepare the return frame
                        tx_header = FrameHeader(container_type=CT_RETURN,
                                                coder_type=CODE_MARSHAL,
                                                transaction_id=rx_header.transaction_id,
                                                address_size=address_size)
                        tx_frame = Frame(tx_header, back)
                        await asyncio.sleep(0.001)  # this allows other threads to be executed in parallel
                        # put it in the write queue
                        await send_msg(writer, tx_frame.encode())
                        if DEBUG: print('tx_frame send to client:', tx_frame)

            except Exception as e:
                e_repr = repr(e)
                tb = ''.join(traceback.format_exc())
                # Catch read exceptions and give feedback
                self._write_log('Client RX-Loop: Exception:')
                self._write_log(e_repr)
                self._write_log(repr(tb))
                # exceptions are send directly not via writer loop to the client
                if rx_header is None:
                    # no header received we give unclear feedback
                    try:
                        await send_msg(writer,
                                       Frame(FrameHeader(CT_EXCEPTION, CODE_MARSHAL), repr(e), repr(tb)).encode())
                    except:
                        self._write_log('LOOP read_from_client(): Unspecific Exception not send to client; '
                                        'sub-exception traceback: %s' % ''.join(traceback.format_exc()))
                else:
                    # header received we give specific feedback
                    try:

                        await send_msg(writer,
                                       Frame(FrameHeader(CT_EXCEPTION, CODE_MARSHAL,
                                                         transaction_id=rx_header.transaction_id,
                                                         command_id=rx_header.command_id,
                                                         channel_id=rx_header.channel_id,
                                                         client_id=rx_header.client_id,
                                                         counter=rx_header.counter),
                                             repr(e), tb).encode())
                    except:
                        self._write_log('LOOP Client: %s: Exception return not send to client; '
                                        'sub-exception traceback: %s' % (client.name, ''.join(traceback.format_exc())))

        self._write_log('Client %s read from client loop/task stopped' % client.name)
        # if this loop is stopped the client is stopped or died
        # we delete the client
        client_id = client.id
        self._write_log('Client %i died' % client_id)
        try:
            self.rpc_delete_client(client, client.control_key)
        except:
            print('Client delete failed:', traceback.format_exc())
            pass

    def _get_writer_bufferlen(self, writer):
        """
        helper delivering the StreamWriter buffer length
        Used for load statistics
        :param writer: StreamWriter object to get the length from
        :return:
        """
        return len(writer.transport._buffer)

    def __get_new_transaction(self, tto=None):
        '''
        a helper function to create a new transaction object
        :param tto: give the transaction_timeout for this transaction
        :return:
        '''
        if tto is None:
            tto = TTO_DEFAULT
        return Transaction(self._loop, self._server_transactions, self._server_transaction_id_hdl, self._address_size, tto,
                           TTO_DEFAULT)

    def _set_channel_slot_times(self):
        """
        helper setting the channel slot explicit
        :return:
        """
        l = len(self._channels) / 2
        if l == 0:
            return
        current_slot_delta_time = self._max_slottime / l
        if current_slot_delta_time < self._min_slottime:
            current_slot_delta_time = self._min_slottime
        for k, c in self._channels.items():
            if type(k) is not int:
                continue
            c.set_slot_time(current_slot_delta_time)

    def _execute_rpc_threaded(self, client, rx_bytes):
        """
        rpc commands on the server are executed in extra thread so that we must not wait in the client read loops
        for the return of the commands
        :param client: client
        :param rx_bytes: bytes that contains the received command content
        :return:
        """
        loop = asyncio.new_event_loop()
        asyncio.set_event_loop(loop)
        rx_header = FrameHeader(CT_DECODE, rx_bytes, address_size=self._address_size)
        rx_payload = rx_bytes[len(rx_header):]
        # server rpc calls are executed here
        try:
            srv_cmd = self._rpc_methods[rx_header.command_id]
        except IndexError:
            srv_cmd = None
        if srv_cmd is None or srv_cmd == self.rpc_authenticate_client:
            # the authentication command (cmd_id==0) is protected and is used only in the client_connect()
            raise StreamError_Command('StreamServer - RPC no/unknown command given')
        # prepare the command arguments
        rx_frame = Frame(rx_header)
        rx_frame.decode_payload(rx_payload)
        # we replace the client argument with the client object
        args = rx_frame.args
        try:
            l = len(args)
        except:
            l = 0
        if l > 0 and (type(args) not in {str, bytes}):
            args = [client] + list(args[1:])
        else:
            args = [client]
        kwargs = rx_frame.kwargs
        if 'client_id' in kwargs:
            del kwargs['client_id']
        # execute the command
        back = srv_cmd(*args, **kwargs)
        # prepare the return frame
        tx_header = FrameHeader(container_type=CT_RETURN,
                                coder_type=CODE_PICKLE,
                                transaction_id=rx_header.transaction_id,
                                address_size=self._address_size)
        # if (hasattr(back, '__iter__')) and (type(back) not in {str,bytes}):
        #    tx_frame = Frame(tx_header, *back)
        # else:
        tx_frame = Frame(tx_header, back)
        # put it in the write queue
        loop.run_until_complete(send_msg(client._async_writer, tx_frame.encode()))
        if DEBUG: print('tx_frame send to client:', tx_frame)

    def close_server(self):
        # Close the server
        if self._server is not None:
            self._server.close()
            try:
                loop = asyncio.get_event_loop()
                loop.run_until_complete(self._server.wait_closed())
                loop.close()
            except:
                pass



    # RPC (Remote Procedure Calls)

    #  rpc methods (exchange via channel 0)
    # the first argument for server rpc commands must be always the client_id this argument will be replaced
    # in the call with the client object so that it is available during the execution

    def rpc_authenticate_client(self, authentication_key, client_name, rpc_methods_descs, raw_key=None):
        '''
        This command is an exception and cannot be called from the client explicitly
        it is just used in the authentication process during the client_connect()
        Note:: Direct usage will lead into an exception!
        :param authentication_key:
        :param client_name:
        :param rpc_methods_descs:
        :param raw_key:
        :return:
        '''
        if authentication_key != raw_key:
            return 'Wrong authentication_key given'
        if DEBUG: print('Client authentication successful')
        return 0, client_name, rpc_methods_descs

    def rpc_echo(self, client_id, *args,**kwargs):
        """
        With this method one can test the rpc connection
        :param client_id: first parameter on server rpc commands is always the client id
                          (will be replaced by client object after frame is received)
        :param args: arguments list to be echoed into the return
        :param kwargs:  keyword arguments dict to be echoed into the return

        :return args,kwargs
        """
        return args,kwargs

    def rpc_create_new_channel(self, client_id, chl_name, public=True, lifetime=None, single_in=True,
                               fixed_slot_time=None):
        """
        This command creates a new channel
        :param client_id: first parameter on server rpc commands is always the client id
                          (will be replaced by client object after frame is received)
        :param chl_name: New channels name
        :param public: Is the channel public or not?
                       * True - is public channel will be shown in channel list
                       * False - channel is not shown in channel list and can only be reached if the name or
                                 ID is known by the client
        :param lifetime: We can create temporary channels too they will automatically deleted after lifetime
        :param single_in: Can the channel have other writers?
                          * True - only owner can write
                          * False - other channel can subscribe for writing
        :param fixed_slot_time: If the parameter is set a fixed slot time in seconds is set.
                                A part of load balancing is switched off in this case.
                                Special channel might be prioritized higher by this parameter
        :return: channel id, channel control key
        """
        client = client_id  # the client object is here already in (see read loop)
        self._write_log('New channel %s request from client %i/%s' % (chl_name, client.id, client.name))
        # the given channel name must be stored as a string (what ever we have received)
        chl_name = str(chl_name)

        if chl_name in self._channels or chl_name == self._statistic_channel_name:
            self._write_log('New channel %s request from client %i/%s - denied; channel exists already' % (
                chl_name, client.id, client.name))
            raise StreamError_Channel('Channel name=%s already in use' % repr(chl_name))
        chl_id = self._channel_id_hdl.get_new_id()
        if chl_id is None:
            self._write_log('New channel %s request from client %i/%s - denied; no free channel' % (
                chl_name, client.id, client.name))
            raise StreamError_Channel('No free channel')
        new_chl = ServeChannel(chl_id, chl_name, client.id, public, single_in, lifetime,
                               self._loop, self._address_size, fixed_slot_time)
        if DEBUG: print('New channel object created', new_chl)
        # we fill the channel dict with both id and name (channel can be accessed via name or id)
        self._channels[chl_id] = new_chl
        self._channels[chl_name] = new_chl
        self._set_channel_slot_times()
        client.register_owned_chl(new_chl)
        new_chl.add_client_writer(client)
        self._write_log('Channel %s creation finished' % chl_name)
        if DEBUG: print('Channel data returned:', chl_id, new_chl.control_key)
        if lifetime is not None:
            #start timer to delete the channel:
            threading.Timer(lifetime,self.rpc_delete_channel,args=('SERVER', new_chl.id, new_chl._control_key))
        return chl_id, new_chl.control_key

    def rpc_subscribe_read_from_channel(self, client_id, chl_id_or_name, buffer_size=100,
                                        full_behavior=RING_BUFFER_FULL, fill_rate=1):
        """
        rpc method to subscribe a read on a channel
        Based on the given parameters a read buffer is created to read the data
        :param client_id: first parameter on server rpc commands is always the client id
                          (will be replaced by client object after frame is received)
        :param chl_id_or_name: channel identification id or name
        :param buffer_size: size of the read buffer
        :param full_behavior: Type of full behavior see Buffer object for details
        :param fill_rate: Integer defines how often we sample,
                          * if fill rate is 1 any channel item is put in the read buffer
                          * if fill rate is 2 any second item will be put in the read buffer
                          * if fill rate is n and n item will be put in the read buffer
        return channel name, channel id
        """
        client = client_id  # the client object is here already in (see read loop)
        self._write_log('Channel read subscription request from client %i/%s' % (client.id, client.name))
        try:
            chl = self._channels[chl_id_or_name]
        except KeyError:
            raise StreamError_Channel('Given channel name=%s is unknown' % repr(chl_id_or_name))
        new_read_buffer = Buffer(buffer_size, full_behavior, fill_rate=fill_rate,
                                 chl_id=chl.id, chl_name=chl.name, client=client)
        chl.add_client_reader(new_read_buffer)
        client._read_bufs.add(new_read_buffer)
        self._write_log('Channel %s(%s) read subscription finished' % (chl.name, chl.id))
        return chl.name, chl.id

    def rpc_unsubscribe_read_from_channel(self, client_id, chl_id_or_name):
        """
        Delete the subscription of a read buffer for a client on the channel
        :param client_id: first parameter on server rpc commands is always the client id
                          (will be replaced by client object after frame is received)
        :param chl_id_or_name: channel identification id or name
        """
        client = client_id  # the client object is here already in (see read loop)
        self._write_log('Channel read subscription delete request from client %i/%s' % (client.id, client.name))
        try:
            chl = self._channels[chl_id_or_name]
        except KeyError:
            raise StreamError_Channel('Given channel name=%s is unknown' % repr(chl_id_or_name))
        found = False
        client_id=client.id
        rb=chl.remove_client_reader(client.id)
        client._read_bufs.remove(rb)
        del rb
        self._write_log('Channel read subscription deleted')
        return True

    def rpc_subscribe_write_to_channel(self, client_id, chl_id_or_name):
        """
        subscribe on a channel as additional writer
        Only for channels which are configured as multi input (no single_in)
        :param client_id: first parameter on server rpc commands is always the client id
                          (will be replaced by client object after frame is received)
        :param chl_id_or_name: channel identification id or name
        :return: channel_name,channel_id
        """
        client = client_id  # the client object is here already in (see read loop)
        self._write_log('Channel write subscription request from client %i/%s' % (client.id, client.name))
        try:
            chl = self._channels[chl_id_or_name]
        except KeyError:
            raise StreamError_Channel('Given channel name=%s is unknown' % repr(chl_id_or_name))
        if chl.is_single_in:
            raise StreamError_Channel('Given channel name=%s; rejected other writers' % repr(chl_id_or_name))
        client._write_chls[chl.id] = chl
        client._write_chls[chl.name] = chl
        chl.add_client_writer(client)
        self._write_log('Channel write subscription finished')
        return chl.name, chl.id

    def rpc_unsubscribe_write_to_channel(self, client_id, chl_id_or_name):
        """
        Delete the subscribe on a channel as additional writer
        :param client_id: first parameter on server rpc commands is always the client id
                          (will be replaced by client object after frame is received)
        :param chl_id_or_name: channel identification id or name
        :return: True
        """
        client = client_id  # the client object is here already in (see read loop)
        self._write_log('Channel read subscription request from client %i/%s' % (client.id, client.name))
        try:
            chl = self._channels[chl_id_or_name]
        except KeyError:
            raise StreamError_Channel('Given channel name=%s is unknown' % repr(chl_id_or_name))
        if chl.is_single_in:
            raise StreamError_Channel('Given channel name=%s; rejected other writers' % repr(chl_id_or_name))
        client._write_chls[chl.id] = chl
        client._write_chls[chl.name] = chl
        chl.remove_client_writer(client.id)
        return True

    def rpc_delete_channel(self, client_id, chl_id_or_name, channel_control_key):
        """
        rpc method to delete a channel
        can only be called by the owner
        :param client_id: first parameter on server rpc commands is always the client id
                          (will be replaced by client object after frame is received)
        :param chl_id_or_name: channel identification id or name
        :param channel_control_key:  key used for authenticate the deletion
        :return:
        """
        client = client_id  # the client object is here already in (see read loop)
        self._write_log('Delete channel %i request received' % (chl_id_or_name))
        chl = self._channels.get(chl_id_or_name)
        if chl is None:
            raise StreamError_Channel('Unknown channel %s' % repr(chl_id_or_name))
        chl.stop_channel()  # stop background thread
        chl_id = chl.id
        chl_name = chl.name
        if client == 'SERVER':
            #Server request the delete!
            client=self._clients[chl._owner]
        if chl_id not in client._owned_chl_ids:
            self._write_log('Delete channel %i request denied, client is not owner' % (chl_id))
            raise StreamError_Authenticate('Authenticate failed, client not owner of the channel')
        if chl.control_key != channel_control_key:
            self._write_log('Delete channel %i request denied, wrong channel control key given' % (chl_id))
            raise StreamError_Authenticate('Authenticate failed, wrong channel control key given')
        # remove all readers in all clients first
        for sub_client in self._clients.values():
            for buf in chl._read_bufs:
                if buf in sub_client._read_bufs:
                    sub_client._read_bufs.remove(buf)
            if chl_id in sub_client._write_chls:
                del sub_client._write_chls[chl_id]
                del sub_client._write_chls[chl_name]
        # remove the channel from all dicts
        client._owned_chl_ids.remove(chl_id)
        del self._channels[chl_id]
        del self._channels[chl_name]
        self._channel_id_hdl.free_id(chl_id)
        self._set_channel_slot_times()
        self._write_log('Channel %i/%s deleted' % (chl_id, chl_name))
        return 'Channel deleted'

    def rpc_delete_client(self, client_id, client_control_key):
        """
        rpc command to delete a client and close the connection
        :param client_id: first parameter on server rpc commands is always the client id
                          (will be replaced by client object after frame is received)
        :param client_control_key: key used for authenticate the deletion
        :return:
        """
        client = client_id  # the client object is here already in (see read loop)
        self._write_log('Delete client (%i,%s) request received' % (client.id, client.name))
        # first we check the control key:
        if client.control_key != client_control_key:
            raise StreamError_Authenticate('Authenticate failed, client cannot be deleted')
        with self._srv_lock:
            for chl_id in list(client._owned_chl_ids):
                chl = self._channels[chl_id]
                self.rpc_delete_channel(client, chl.id, chl.control_key)
            # finally remove the two client entries in the clients dict
            try:
                del self._clients[client.id]
            except:
                pass
            try:
                del self._clients[client.name]
            except:
                pass
            self._client_id_hdl.free_id(client.id)
            self._write_log('Client (%i,%s) removed' % (client.id, client.name))
            return 'Client removed'

    def rpc_get_clients(self, client_id):
        """
        delivers a list off all connected clients on this server
        :param client_id: first parameter on server rpc commands is always the client id
                          (will be replaced by client object after frame is received)
        :return: list of tuples (client name, client id)
        """
        return [(client.name, client.id) for k, client in self._clients.items() if type(k) is int]

    def rpc_get_client_id(self, client_id,client_name):
        """
        Translates the given client name into the client id
        (If id is given id will be given back any way)
        :param client_id: first parameter on server rpc commands is always the client id
                          (will be replaced by client object after frame is received)
        :param client_name: client name from the client from which the related id should be delivered
        :return:
        """
        client = client_id  # the client object is here already in (see read loop)
        if client_name not in self._clients:
            raise StreamError_Target('Given target client (%s) unknown!' % str(client_name))
        return self._clients[client_name].id

    def rpc_get_channels(self, client_id):
        """
        Delivers a list of all public channels that exists on the server
        :param client_id: first parameter on server rpc commands is always the client id
                          (will be replaced by client object after frame is received)
        :return: list of tuples (channel name, channel id)
        """
        return [(chl.name, chl.id) for k, chl in self._channels.items() if (type(k) is int) and chl.is_puplic]

    def rpc_get_client_info(self, client_id, target_client_name_or_id):
        """
        Ask for the client info delivers a dict with client information
        id, name and rpc_method description
        :param client_id: first parameter on server rpc commands is always the client id
                          (will be replaced by client object after frame is received)
        :param target_client_name_or_id: client identification name of id
        :return: dict with 'ID','NAME', 'RPC_METHOD_DESC' keys
        """
        client = client_id  # the client object is here already in (see read loop)
        if target_client_name_or_id not in self._clients:
            raise StreamError_Target('Given target client (%s) unknown!' % str(target_client_name_or_id))
        target_client= self._clients[target_client_name_or_id]
        return_dict={'ID':target_client.id,
                     'NAME':target_client.name,
                     'RPC_METHOD_DESC':target_client.get_rpc_method_desc()
                     }
        return return_dict

    def rpc_start_data_transfer_on_channel(self, client_id, chl_id_or_name):
        """
        rpc method to start a data transfer transaction on a channel
        A transaction id is generated that should be used for the data frames following until the end frame is given
        Note:: The transaction isd is a channel specific number and cannot be reused on other channels!
        :param client_id: first parameter on server rpc commands is always the client id
                          (will be replaced by client object after frame is received)
        :param chl_id_or_name: channel identification name of id
        :return: transaction id
        """
        client = client_id  # the client object is here already in (see read loop)
        chl = client._write_chls.get(chl_id_or_name)
        if chl is None:
            raise StreamError_Channel(
                'Channel (%s) not found or client is no writer on this channel' % repr(chl_id_or_name))
        t_id = chl.get_new_transaction_id()
        return t_id

    def rpc_start_statistic_channel(self, client_id, fill_rate=100, fixed_slot_time=None):
        """
        start the statistic channel to track the load of the server
        :param client_id: first parameter on server rpc commands is always the client id
                          (will be replaced by client object after frame is received)
        :param fill_rate: How often the readers should put their statistic information into the statistic channel
        :param fixed_slot_time: Channel might be prioritized higher (not recommended)
        :return: statistc channel name is returned so that a reader might be subscribed by the client
        """
        if self._statistic_chl is None:
            self._statistic_chl_fill_rate = fill_rate

            new_id = self._channel_id_hdl.get_new_id()
            new_chl = ServeChannel(new_id, self._statistic_channel_name, -1, is_public=True, is_single_in=False,
                                   lifetime=None,
                                   loop=self._loop, address_size=self._address_size,
                                   fixed_slot_time=fixed_slot_time
                                   )
            if DEBUG: print('New channel object created', new_chl)
            # we fill the channel dict with both id and name (channel can be accessed via name or id)
            self._channels[new_id] = new_chl
            self._channels[self._statistic_channel_name] = new_chl
            self._set_channel_slot_times()
            self._statistic_chl = new_chl
            for chl in self._channels.values():
                chl.set_statistic_channel(new_chl, fill_rate)
            return self._statistic_channel_name
        else:
            # channel exists already!
            return self._statistic_channel_name

    def rpc_get_statistic_channel_state(self,client_id):
        """
        :param client_id: first parameter on server rpc commands is always the client id
                          (will be replaced by client object after frame is received)

        :return:
                * True, channel name - channel is active
                * False, channel name  -channel is inactive
        """
        if self._statistic_chl is not None:
            return True,self._statistic_channel_name
        else:
            return False,self._statistic_channel_name

    def rpc_stop_statistic_channel(self, client_id):
        """
        stop the statistic channel
        :param client_id: first parameter on server rpc commands is always the client id
                          (will be replaced by client object after frame is received)
        :return:
        """
        if self._statistic_chl is not None:
            statistic_chl = self._statistic_chl
            statistic_header = FrameHeader(CT_DATA_END, CODE_MARSHAL, channel_id=statistic_chl.id)
            statistic_chl.clear_buffers()
            statistic_chl.put_nowait(CT_DATA_END, Frame(statistic_header, 'END_DATA').encode())
            statistic_chl._task_trigger.set()
            # we wait a short moment before deleting the channel
            time.sleep(2)
            for chl in self._channels.values():
                chl.set_statistic_channel(None, 1)
            del self._channels[statistic_chl.name]
            del self._channels[statistic_chl.id]
            self._set_channel_slot_times()
            self._statistic_chl = None
            self._channel_id_hdl.free_id(statistic_chl.id)
            return True
        else:
            return False



"""
This code is taken from PyPi stream_service package
https://github.com/BR1py/stream_service
"""

from __future__ import absolute_import
import marshal
import pickle
import gzip
from collections import OrderedDict
from .helpers import *

# Frame related definitions and classes:

# CONTAINER TYPES:
CT_DATA = 0  # Data frame
CT_RETURN = 0b1  # Return frame for rpc calls
CT_CMD = 0b10  # client related rpc command
CT_SRV_CMD = 0b11  # server related rpc command
CT_EXCEPTION = 0b100  # exception used in case of errors in rpc calls
CT_DATA_END = 0b101  # data stream on channel will be stopped
CT_DECODE = 0b110  # DECODE the given raw data

CONTAINER_TYPE_STR = {CT_DATA: 'DATA', CT_RETURN: 'RETURN', CT_CMD: 'CMD', CT_SRV_CMD: 'SRV_CMD', CT_EXCEPTION:
    'EXCEPTION', CT_DECODE: 'DECODE', CT_DATA_END: 'DATA_END'}

# CODE
CODE_GZIP = 0b1  # zip the bytes after serialization (may help for huge data packages) (This is first bit of the frame!)
CODE_MARSHAL = 0b010  # marshal the given objects for serialization
CODE_PICKLE = 0b100  # pickle the objects for serialization
CODE_STR = {CODE_MARSHAL: 'CODE_MARSHAL', CODE_PICKLE: 'CODE_PICKLE'}

try:
    # optional if numpy can be imported
    import numpy

    NUMPY_FROM_BYTES = numpy.frombuffer
    CODE_NUMPY = 0b110  # serialize the given numpy array (we use numpy serializer to_bytes/from_bytes)
    CODE_STR[CODE_NUMPY] = 'NUMPY'
except:
    CODE_NUMPY = None


    def dummy(*args, **kwargs):
        pass


    NUMPY_FROM_BYTES = dummy

# Data Frame mark end of transmissions via special counter values:
DATA_END_COUNTER = 0


class FrameHeader():
    __slots__ = (
        '_container_type', '_coder_type', '_transaction_id', '_command_id', '_channel_id', '_client_id', '_counter',
        '_address_size', '_transaction_id_bytes', '_command_id_bytes', '_channel_id_bytes', '_client_id_bytes',
        '_counter_bytes', '_id_bits', '_header_len', '_sub_header', '_address_mask', '_max_counter', '_raw_bytes',
    )

    # The order here is important for quicker, easier decoding of information in the server
    BIT_ITEM_RELATION = OrderedDict([(0b1, '_channel_id'),
                                     (0b10, '_client_id'),
                                     (0b100, '_command_id'),
                                     (0b1000, '_transaction_id'),
                                     (0b10000, '_counter')])

    def __init__(self,
                 container_type=CT_DATA,
                 coder_type=CODE_MARSHAL,
                 channel_id=None,
                 command_id=None,
                 client_id=None,
                 transaction_id=None,
                 counter=None,
                 address_size=2):
        '''
        FrameHeader object contains all data that are put in the header of the frame.
        The header does not have a fixed length instead the length is implicit calculated by the bits
        set in the second byte of the header (defines which ids/counters are active or not)

        :param container_type:  Defines what kind of frame should be generated
        :param coder_type: Define how the data given should be codeed (serialized)

        :param transaction_id: Put transaction ID in the frame (None - header will be shortened no info in)
        :param command_id: put command_id in the header (None - header will be shortened no info in)
        :param channel_id: put channel_id into the header (None - header will be shortened no info in)
        :param client_id: put client_id of the target client into the header (None - header will be shortened no info in)
        :param counter: put a counter into the header (None - header will be shortened no info in)
        :param address_size: The address size used of on the server must be given to encode the ids correctly
        '''
        # fill the self.xxx attributes
        self._container_type = container_type
        self._coder_type = coder_type
        self._transaction_id = transaction_id
        self._command_id = command_id
        self._channel_id = channel_id
        self._client_id = client_id
        self._counter = counter
        self._address_size = address_size

        # addresse related data
        mask = 0
        for i in range(address_size):
            mask = (mask << 8) + 0xFF
        self._address_mask = mask
        self._max_counter = 2 ** (address_size * 8)
        if container_type == CT_DECODE:
            # In this special case we expect the raw_bytes in the first argument after the container type!
            # It's the code parameter
            # or the header is initialized without data (empty) might be decoded later on
            if coder_type != CODE_MARSHAL:
                self.decode(coder_type)
        else:
            if coder_type not in CODE_STR:
                raise ValueError('Given coder_type is unknown')
            self._id_bits = 0
            # fill in the id_byte attributes and calculate header length
            l = 2
            for (bit, name) in self.BIT_ITEM_RELATION.items():
                arg = self.__getattribute__(name)
                if arg is not None:
                    self._id_bits = self._id_bits | bit
                    l += address_size
            self._header_len = l
            self._raw_bytes = self.__build_raw_bytes()

    def __build_raw_bytes(self):
        address_size = self._address_size
        mask = self._address_mask
        a = ((self._coder_type & 0b1111) << 4) + (self._container_type & 0b1111)
        first_bytes = a.to_bytes(1, BO) + self._id_bits.to_bytes(1, BO)
        items = [first_bytes]+[int_address_to_bytes(self.__getattribute__(v), address_size, mask) for v in
                 self.BIT_ITEM_RELATION.values() if self.__getattribute__(v) is not None]
        return b''.join(items)

    def __set_id_attr(self, bit, value):
        name = self.BIT_ITEM_RELATION[bit]
        if value is None:
            if self.__getattribute__(name) is not None:
                if self._id_bits & bit != 0:  # normally this should be always the case
                    self._id_bits = self._id_bits ^ bit
                    self._header_len -= self._address_size
                self.__setattr__(name, None)
        else:
            if self.__getattribute__(name) is None:
                self._id_bits = self._id_bits | bit
                self._header_len += self._address_size
            self.__setattr__(name, value)
        self._raw_bytes = self.__build_raw_bytes()

    # make some internals as properties available
    @property
    def container_type(self):
        return self._container_type

    @property
    def coder_type(self):
        return self._coder_type

    @property
    def address_size(self):
        return self._address_size

    @property
    def transaction_id(self):
        return self._transaction_id

    @property
    def command_id(self):
        return self._command_id

    @property
    def channel_id(self):
        return self._channel_id

    @property
    def client_id(self):
        return self._client_id

    @property
    def counter(self):
        return self._counter

    def set_transaction_id(self, value):
        '''
        adapt the transaction id in the header to a new value
        :param value: id to be put in (Give None to delete)
        :return:
        '''
        bit = 0b1000
        return self.__set_id_attr(bit, value)

    def set_command_id(self, value):
        '''
        adapt the command id
        :param value: id to be put in (Give None to delete)
        :return:
        '''
        bit = 0b100
        return self.__set_id_attr(bit, value)

    def set_channel_id(self, value):
        '''
        adapt the channel id
        :param value: id to be put in (Give None to delete)
        :return:
        '''
        bit = 0b1
        return self.__set_id_attr(bit, value)

    def set_client_id(self, value):
        '''
        adapt the client id
        :param value: id to be put in (Give None to delete)
        :return:
        '''
        bit = 0b10
        return self.__set_id_attr(bit, value)

    def set_counter(self, value):
        '''
        adapt the counter
        :param value: id to be put in (Give None to delete)
        :return:
        '''
        bit = 0b10000
        return self.__set_id_attr(bit, value)

    def count_up(self):
        if self._counter is None:
            raise ValueError('No counter value defined')
        self._counter += 1
        if self._counter > self._max_counter:
            self._counter = 1  # we count from one because counter==0 is DATA_END_COUNTER!
        address_size = self._address_size
        # for the raw bytes we do it quick (counter is always in the last bytes
        self._raw_bytes = self._raw_bytes[:-address_size] + int_address_to_bytes(self._counter, address_size)

    def encode(self, tmp_client_id=None, cnt_up=False):
        if cnt_up:
            self.count_up()
        if tmp_client_id is not None:
            old_client_id = self.client_id
            self.set_client_id(tmp_client_id)
            data = self.__build_raw_bytes()
            self.set_client_id(old_client_id)
        else:
            data = self._raw_bytes
        return data

    def decode(self, received_data):
        if received_data is None:
            return
        sub_header = received_data[:2]
        self._container_type = c1 = sub_header[0] & 0b1111
        self._coder_type = c2 = (sub_header[0] & 0b11110000) >> 4
        start = 2
        end = 2
        l = 2
        self._id_bits = sub_header[1]
        for bit, name in self.BIT_ITEM_RELATION.items():
            if self._id_bits & bit != 0:
                end = start + self._address_size
                d = received_data[start:end]
                self.__setattr__(name, int.from_bytes(d, BO))
                start = end
                l += self._address_size
            else:
                self.__setattr__(name, None)
        self._header_len = l
        self._raw_bytes = self.__build_raw_bytes()

    def __repr__(self):
        out = ['FrameHeader(',
               ' container_type = %s,' % CONTAINER_TYPE_STR[self._container_type]]
        coder_type_str = CODE_STR[self._coder_type & 0b1110]
        if self._coder_type & 0b1 != 0:
            coder_type_str = 'CODE_GZIP/' + coder_type_str
        out.append(' coder_type = %s,' % coder_type_str)
        for bit, name in self.BIT_ITEM_RELATION.items():
            d = self.__getattribute__(name)
            if d is not None:
                out.append(' %s = %i,' % (name, d))
        out[-1] = out[-1][:-1] + ')'
        return ''.join(out)

    def __len__(self):
        return self._header_len


# container classes

class Frame():
    __slots__ = ('_header', '_args', '_kwargs')

    def __init__(self, header, *args, **kwargs):
        '''
        The Frame is the abstraction for the data send out to/from the clients
        :param header: FrameHeader object based on this object we decided how to code the given data
        :param args: objects to be codeed
        :param kwargs:objects to be codeed as named arguments
        '''
        if header.container_type == CT_DECODE:
            if len(args) != 1 and len(kwargs) != 0:
                raise ValueError('In decode mode only a bytes argument is accepted')
            if len(args) > 0:
                raw_data = args[0]
                self.decode(raw_data)
            else:
                self._args = tuple()
                self._kwargs = {}

        else:
            self._header = header  # first argument must be the header!
            self._args = args
            self._kwargs = kwargs

    # make some internal data available

    @property
    def header(self):
        return self._header

    @property
    def args(self):
        return self._args

    @property
    def kwargs(self):
        return self._kwargs

    def set_arguments(self, *args, **kwargs):
        '''
        reset the arguments of the Frame object
        (reset the payload relevant data)
        :param args:
        :param kwargs:
        :return:
        '''
        self._args = args
        self._kwargs = kwargs

    def encode(self, client_id=None, cnt_up=False):
        '''
        Encode the Frame into bytes
        :param client_id: optional client_id might be temporary exchanged in the header just for this sending
                          (used on server to exchange target and source client)
        '''
        header_bytes = self._header.encode(client_id, cnt_up)
        coder_type = self._header.coder_type
        if self._args is None and self._kwargs is None:
            return header_bytes
        if len(self._args) == 0 and len(self._kwargs) == 0:
            return header_bytes

        payload_bytes = b''
        if coder_type & 0b110 & CODE_MARSHAL != 0:
            payload_bytes = marshal.dumps((self._args, self.kwargs))
        elif coder_type & 0b110 & CODE_PICKLE != 0:
            payload_bytes = pickle.dumps((self._args, self.kwargs))
        elif coder_type & 0b110 & CODE_NUMPY != 0:
            data = ([(arg.to_bytes(), arg.shape) for arg in self._args],
                    [(k, arg.to_bytes(), arg.shape) for k, arg in self._kwargs.items()])
            payload_bytes = marshal.dumps(data)
        if coder_type & CODE_GZIP != 0:
            payload_bytes = gzip.compress(payload_bytes)
        return header_bytes + payload_bytes

    def decode_header_update(self, receive_bytes, address_size=2):
        '''
        Decode the header based on the received bytes and redefine the Frame header
        Note:: The Frame arguments will be resetted to empty (default)
        :param receive_data:
        :param address_size:
        :return:
        '''
        header = FrameHeader(CT_DECODE, address_size=address_size)
        header.decode(receive_bytes)
        self._header = header
        self._args = tuple()
        self._kwargs = {}

    def decode_payload(self, payload_bytes):
        '''
        Decode the payload given based on the Frame header definition
        :param payload_bytes: payload bytes to be decoded
        :return:
        '''
        coder_type = self._header.coder_type
        self._args = tuple()
        self._kwargs = {}
        data = (tuple(), {})
        if len(payload_bytes) == 0 or payload_bytes is None:
            return
        if coder_type & CODE_GZIP != 0:
            payload_bytes = gzip.decompress(payload_bytes)
        if coder_type & 0b110 & CODE_MARSHAL != 0:
            data = marshal.loads(payload_bytes)
        elif coder_type & 0b110 & CODE_PICKLE != 0:
            data = pickle.loads(payload_bytes)
        elif coder_type & 0b110 & CODE_NUMPY != 0:
            data = marshal.loads(payload_bytes)
            args = [NUMPY_FROM_BYTES(i[0], i[1]) for i in data[0]]
            kwargs = dict([(i[0], NUMPY_FROM_BYTES(i[1], i[2])) for i in data[1]])
            data = (args, kwargs)
        # sometimes we get here something like (a,b,c) or (a,) we do ot get the generator (a)!
        # if (a,) we extract a only we guess this is always a single value to extract -> but this might be wrong!?
        # args=
        # if type(args) is tuple and len(args)==1:
        #    args=args[0]
        # if self._header.container_type==CT_RETURN:
        # here I'not sure here have some troubles with returns values ...
        self._args = data[0]
        self._kwargs = data[1]

    def _get_args(self, *args, **kwargs):
        return (locals().values())

    def decode(self, receive_data, address_size=2):
        '''
        Decode the received data and redefine the Frame content
        :param receive_data:
        :param address_size:
        :return:
        '''
        self.decode_header_update(receive_data, address_size)
        l = len(self._header)
        return self.decode_payload(receive_data[l:])

    def __repr__(self):
        out = ['Frame(', ' %s,' % repr(self.header)]
        for v in self._args:
            out.append(' %s,' % (repr(v)))
        for k, v in self._kwargs:
            out.append(' %s = %s,' % (repr(k), repr(v)))
        out[-1] = out[-1][:-1] + ')'
        return ''.join(out)




"""
This code is taken from PyPi stream_service package
https://github.com/BR1py/stream_service
"""

import sys
import os
import time
import ssl

if "-root_path" in sys.argv: #started via shell script
    ROOT=os.path.dirname(os.path.dirname(os.path.normpath(sys.argv[(sys.argv.index("-root_path")+1)])))
else:
    ROOT=os.path.dirname(os.path.dirname(__file__))
if ROOT not in sys.path:
    sys.path.append(ROOT)

from stream_service import StreamChannelServer_Process

if __name__ == '__main__':
    srv = StreamChannelServer_Process(ip_address='127.0.0.1',client_authenticate=b'STRM_Client')
    srv.start()
    # let's wait until the server might be finished
    srv.join()



"""
This code is taken from PyPi stream_service package
https://github.com/BR1py/stream_service
"""

from __future__ import absolute_import
from .helpers import *
from . import frame
from . import rpc
from . import buffer



"""
In this example we show how we can create two clients using the class StreamChannelClient_Thread as a super class.
We extend the base functionality by adding rpc-methods (remote procedure calls) that can be executed and used by
other clients.
create 4 Clients that will connect to the server and setup several channels
Then different type of data is streamed in between the clients


    Client1
    call rpc Client2
            \
             \
              \
               Server
              /
             /
            /
    Client2
    call rpc Client 1

REMARK: To let the example run start the run_server.py script in a parallel process
"""

import asyncio
import sys
import os
import time
import threading
import ssl

if "-root_path" in sys.argv: #started via shell script
    ROOT=os.path.dirname(os.path.dirname(os.path.normpath(sys.argv[(sys.argv.index("-root_path")+1)])))
else:
    ROOT=os.path.dirname(os.path.dirname(__file__))
if ROOT not in sys.path:
    sys.path.append(ROOT)

from stream_service import *

class RPCClass():
    """
    helper class that contains more rpc methods
    """
    def rpc_add(self,a,b):
        return self.myadd(a,b)

    def rpc_substract(self, a, b):
        return a - b

    def myadd(self,a,b):
        return a + b

# 1. Define new client classes with extended rpc function using the StreamChannelClient_Thread as super class

class Client1(StreamChannelClient_Thread):

    def __init__(self,*args,**kwargs):
        super().__init__(*args, **kwargs)
        self.rpc_calc=RPCClass()
        self._rpc_methods_descs, self._rpc_methods = rpc.RPCFactory.class_init_rpc_functions(self, first_items=[
            'rpc_get_rpc_info', 'rpc_activate_write_channel'],address_size=self.address_size)

class Client2(StreamChannelClient_Thread):

    def rpc_multiply(self, a, b):
        return a * b

    def rpc_divide(self, a, b):
        return a / b


if __name__ == '__main__':

    # 2. create the client threads based on the two new class definitions

    clt1=Client1('RPCClient1','127.0.0.1',server_authenticate=b'STRM_Client')
    clt1.start()
    clt1.wait_until_alive()

    print('Client1 connected to server')

    clt2 = Client2('RPCClient2','127.0.0.1', server_authenticate=b'STRM_Client')
    clt2.start()
    clt2.wait_until_alive()
    print('Client2 connected to server')

    # 3. get the related client RPCFactory objects
    rpc_client2= clt1.create_rpc_client_service('RPCClient2')
    rpc_client1 =clt2.create_rpc_client_service('RPCClient1')

    # 4. call the rpc methods on the clients:
    print('Result multiply: 2*3 =',rpc_client2.multiply(2,3))
    print('Result divide: 9/3 =',rpc_client2.divide(9, 3))
    print('Result add: 2+3 =',rpc_client1.calc.add(2,3))
    print('Result substract: 2-3 =',rpc_client1.calc.substract(2, 3))

    # 5. stop the clients
    clt1.stop_client()
    clt2.stop_client()



"""
This code is taken from PyPi stream_service package
https://github.com/BR1py/stream_service

We publish here all the relevant classes for Users
"""
from __future__ import absolute_import
from .server import StreamChannelServer_Process
from .client import StreamChannelClient_Thread
from .lib import *
from .lib.frame import *
from .lib.buffer import RING_BUFFER_FULL,RAISE_BUFFER_FULL,SKIP_BUFFER_FULL,CLEAR_BUFFER_FULL,WAIT_BUFFER_FULL




from setuptools import setup

setup(
    name='stream_service',
    version='0.1.0',
    packages=['stream_service', 'stream_service.lib', 'stream_service.examples'],
    package_dir={'': 'src'},
    url='https://github.com/BR1py/stream_service',
    license='MIT',
    author='B_R',
    author_email='br_development@posteo.org',
    description='Client/Server implementation for data streaming in channels based on python >= 3.5'
)
